"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ApiParser = void 0;
const fs = require("fs");
const path = require("path");
const ts = require("typescript");
const transformer_1 = require("../transformer");
const CompilationUtil_1 = require("./CompilationUtil");
const HandlerTree_1 = require("../transformer/HandlerTree");
const CommandUtil_1 = require("../command/CommandUtil");
class ApiParser {
    async parseApi(options, transformerArgs = {}) {
        if (!fs.existsSync(options.rootDir)) {
            throw new Error(`File does not exist: ${options.rootDir}`);
        }
        const resolvedRootDir = path.resolve(process.cwd(), options.rootDir);
        const rootDirStat = fs.lstatSync(resolvedRootDir);
        options.isDir = rootDirStat.isDirectory();
        const transformers = [
            program => (0, transformer_1.default)(program, {
                ...transformerArgs,
                onTreeExtracted: (err, tree) => this.onTreeExtracted(tree)
            }),
        ];
        const hasTsConfig = !!options.tsconfig;
        if (!hasTsConfig) {
            options.tsconfig = ApiParser.DEFAULT_TSCONFIG;
        }
        let compilationResult = null;
        let tsConfig;
        let tsConfigPath;
        if (options.isDir) {
            const loadedConfig = this.loadTsConfig(options.tsconfig, resolvedRootDir);
            tsConfig = loadedConfig.tsConfig;
            tsConfigPath = loadedConfig.path;
            compilationResult = (0, CompilationUtil_1.compileSources)(tsConfig.fileNames, {
                ...tsConfig.options,
                noEmit: true,
            }, transformers);
        }
        else {
            const loadedConfig = this.loadTsConfig(options.tsconfig, path.dirname(resolvedRootDir));
            tsConfig = loadedConfig.tsConfig;
            tsConfigPath = loadedConfig.path;
            compilationResult = (0, CompilationUtil_1.compileSources)([options.rootDir], {
                ...tsConfig.options,
                noEmit: true,
            }, transformers);
        }
        return {
            tsConfig,
            tsConfigPath,
            compilationResult,
            tree: this.tree,
            programInfo: this.loadApiInfo(options.apiInfo),
        };
    }
    onTreeExtracted(tree) {
        this.tree = (0, HandlerTree_1.MergeHandlerTreeRoots)(this.tree, tree);
    }
    loadTsConfig(tsConfig, resolvedRootDir) {
        let config;
        if (tsConfig) {
            config = tsConfig;
        }
        else {
            config = ApiParser.DEFAULT_TSCONFIG;
        }
        const tsconfigPath = path.resolve(process.cwd(), config);
        if (!fs.existsSync(tsconfigPath)) {
            if (tsConfig) {
                throw new Error(`tsconfig does not exist: ${tsconfigPath}`);
            }
            return {
                tsConfig: {
                    options: ts.getDefaultCompilerOptions(),
                    errors: [],
                    fileNames: []
                },
                path: tsconfigPath,
            };
        }
        return {
            tsConfig: (0, CompilationUtil_1.parseTsConfig)(path.dirname(tsconfigPath), tsconfigPath),
            path: tsconfigPath,
        };
    }
    loadApiInfo(apiInfoPath) {
        if (typeof apiInfoPath !== 'string') {
            return undefined;
        }
        const file = path.resolve(process.cwd(), apiInfoPath);
        if (!fs.existsSync(file)) {
            return undefined;
        }
        const infoBase = require(file);
        return {
            version: infoBase.version,
            title: infoBase.name,
            description: infoBase.description,
            contact: (0, CommandUtil_1.getPackageJsonAuthor)(infoBase),
            license: [
                {
                    name: infoBase.license,
                }
            ],
            homepage: infoBase.homepage,
            basePath: infoBase.basePath,
            host: infoBase.host,
        };
    }
    getTree() {
        return this.tree || null;
    }
}
exports.ApiParser = ApiParser;
ApiParser.DEFAULT_TSCONFIG = 'tsconfig.json';
//# sourceMappingURL=ApiParser.js.map