"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ApiDependency = exports.ApiDependencyCollection = void 0;
const dependency_graph_1 = require("dependency-graph");
const ManagedApiInternal_1 = require("./ManagedApiInternal");
class ApiDependencyCollection {
    constructor() {
        this.DepNo = 0;
        this.dependencyGraph = new dependency_graph_1.DepGraph();
        this.registeredDependencies = new Map();
        this.filledDependencies = new Map();
    }
    instantiateDependency(dependency) {
        if (!this.filledDependencies.has(dependency)) {
            let dependencies = this.walkDependencyGraph(dependency);
            for (const d of dependencies) {
                const filled = this.getFilledDependencies(d.name);
                if (filled.size !== d.numDependencies) {
                    throw new Error('Cannot instantiate dependency: unfilled dependencies');
                }
                const apiDep = this.dependencyGraph.getNodeData(d.name);
                apiDep.instantiate(filled);
                this.addFilledDependency(apiDep);
            }
        }
        return this.filledDependencies.get(dependency).instance;
    }
    walkDependencyGraph(dependency) {
        const depMap = new Map();
        const descrHash = this.getHash(dependency);
        let nodesToWalk = [descrHash];
        do {
            const node = nodesToWalk.pop();
            if (depMap.has(node)) {
                // Already walked this
                continue;
            }
            const nodeData = this.dependencyGraph.getNodeData(node);
            if (!nodeData || this.filledDependencies.has(nodeData.reference)) {
                // Already filled this dependency, or not enough info to fill - skip
                continue;
            }
            const nodeDependencies = this.dependencyGraph.dependenciesOf(node);
            depMap.set(node, nodeDependencies.length);
            nodesToWalk = nodesToWalk.concat(nodeDependencies);
        } while (nodesToWalk.length > 0);
        return Array.from(depMap.entries())
            .map(([name, numDependencies]) => ({ name, numDependencies }))
            .sort((a, b) => a.numDependencies - b.numDependencies);
    }
    getFilledDependencies(name) {
        const depMap = new Map();
        const dependencies = this.dependencyGraph.dependenciesOf(name);
        for (const d of dependencies) {
            const idep = this.dependencyGraph.getNodeData(d);
            if (!this.filledDependencies.has(idep.reference)) {
                throw new Error('Expected dependency to be filled: ' + d);
            }
            const filled = this.filledDependencies.get(idep.reference);
            depMap.set(filled.reference, filled);
        }
        return depMap;
    }
    addFilledDependency(apiDep) {
        this.filledDependencies.set(apiDep.reference, apiDep);
    }
    registerDependency(dep) {
        const hash = this.getHash(dep.reference);
        if (this.dependencyGraph.hasNode(hash)) {
            if (this.dependencyGraph.getNodeData(hash)) {
                throw new Error('Dependency already registered');
            }
            else {
                this.dependencyGraph.setNodeData(hash, dep);
            }
        }
        else {
            this.dependencyGraph.addNode(hash, dep);
        }
        for (const dependency of dep.getDependencies()) {
            const depHash = this.getHash(dependency.dependency);
            if (!this.dependencyGraph.hasNode(depHash)) {
                this.dependencyGraph.addNode(depHash, null);
            }
            this.dependencyGraph.addDependency(hash, depHash);
        }
    }
    getHash(descr) {
        if (!this.registeredDependencies.has(descr)) {
            this.registeredDependencies.set(descr, (this.DepNo++).toString());
        }
        return this.registeredDependencies.get(descr);
    }
}
exports.ApiDependencyCollection = ApiDependencyCollection;
ApiDependencyCollection.HashSepChar = '\0';
class ApiDependency {
    constructor(scope, reference, referenceMode, constructionFunc) {
        this.scope = scope;
        this.reference = reference;
        this.referenceMode = referenceMode;
        this.constructionFunc = constructionFunc;
        this._instance = null;
    }
    static WithConstructor(dependencyClass) {
        return new ApiDependency(null, dependencyClass, "constructor" /* ApiDependencyReferenceMode.Constructor */, a => new dependencyClass(...a));
    }
    static WithFunc(dependencyClass, getterFunc) {
        return new ApiDependency(null, dependencyClass, "function" /* ApiDependencyReferenceMode.Function */, a => getterFunc(...a));
    }
    getDependencies() {
        return [
            ...ManagedApiInternal_1.ManagedApiInternal.GetDependenciesOnConstructor(this.reference),
            ...ManagedApiInternal_1.ManagedApiInternal.GetDependencyParams(this.reference, undefined)
        ];
    }
    get isInstantiated() {
        return this._instance !== null;
    }
    get instance() {
        return this._instance;
    }
    instantiate(dependencies) {
        if (this._instance) {
            return this._instance;
        }
        const dependencyParams = ManagedApiInternal_1.ManagedApiInternal.GetDependencyParams(this.reference, undefined)
            .map(d => dependencies.get(d.dependency).instance);
        this._instance = this.constructionFunc(dependencyParams);
        for (const dep of ManagedApiInternal_1.ManagedApiInternal.GetDependenciesOnConstructor(this.reference)) {
            this._instance[dep.propertyKey] = dependencies.get(dep.dependency).instance;
        }
        return this._instance;
    }
}
exports.ApiDependency = ApiDependency;
//# sourceMappingURL=ApiDependency.js.map