/// <reference types="node" />
import { IApiDefinition, ApiMethod, IApiParamDefinition, IApiDefinitionWithProcessors, IApiProcessors } from "./ApiDefinition";
import { Readable } from "stream";
import { ApiMimeType } from "./MimeTypes";
import { __ApiParamArgs } from "./InternalTypes";
import * as p2r from 'path-to-regexp';
import { ClassConstructor } from "../Util/ClassConstructors";
import { Func, Func0, OptionalAsyncFunc1 } from "../Util/Func";
import { StreamCoercionMode } from "../Util/StreamCoercer";
import * as stream from 'stream';
export type ApiParamsDict = {
    [param: string]: string;
};
export type ApiBodyParamsDict = {
    [param: string]: ApiParamsDict;
};
export type ApiHeadersDict = {
    [paramNameLowercase: string]: string | string[];
};
export type ManagedApiPreInvokeHandlerType<TransportParamsType extends object> = OptionalAsyncFunc1<IApiInvocationContext<TransportParamsType>, IApiInvocationContext<TransportParamsType> | undefined>;
export type ManagedApiPostInvokeHandlerType<TransportParamsType extends object> = OptionalAsyncFunc1<IApiInvocationContextPostInvoke<TransportParamsType>, IApiInvocationResult | undefined>;
export interface IApiBodyContents {
    contentsStream: Readable;
    streamContentsMimeType: ApiMimeType;
    streamContentsMimeRaw: string;
    readStreamToStringAsync: () => Promise<string>;
    readStreamToStringCb: (cb: (err: any, contents?: string) => void) => void;
    parsedBody?: any;
    contentsString?: string;
}
export interface IApiInvocationParams<TransportParamsType extends object = {}> {
    queryParams: ApiParamsDict;
    pathParams: ApiParamsDict;
    headers: ApiHeadersDict;
    bodyContents?: IApiBodyContents;
    bodyParams?: ApiParamsDict;
    transportParams: TransportParamsType;
}
export interface IApiInvocationResult {
    streamMode: StreamCoercionMode;
    statusCode: number;
    body: string | object;
    headers: ApiHeadersDict;
}
export interface IApiInvocationContext<TransportParamsType extends object = {}> {
    apiDefinition: IApiDefinitionWithProcessors<TransportParamsType>;
    invocationParams: IApiInvocationParams<TransportParamsType>;
}
export interface IApiInvocationContextPostInvoke<TransportParamsType extends object = {}> extends IApiInvocationContext<TransportParamsType> {
    result: IApiInvocationResult;
}
export type WrappedApiHandler<TransportParamsType extends object> = (params: IApiInvocationParams<TransportParamsType>) => Promise<IApiInvocationResult>;
export interface RouteTokenKey {
    name: string | number;
    prefix: string;
    suffix: string;
    pattern: string;
    modifier: string;
}
export type RouteToken = string | RouteTokenKey;
export interface IApiHandlerInstance<TransportParamsType extends object> extends IApiDefinition {
    parent: object;
    wrappedHandler: WrappedApiHandler<TransportParamsType>;
    handlerArgs: IApiParamDefinition[];
    routeTokens: RouteToken[];
}
export declare abstract class ManagedApi<TransportParamsType extends object> {
    private readonly useGlobal;
    private static readonly ClsNamespace;
    private static readonly ContextNamespace;
    readonly HookHandlerPreInvoke = "handler-preinvoke";
    readonly HookHandlerPostInvoke = "handler-postinvoke";
    private static apis;
    private static namespace;
    private readonly jsonValidator;
    private classes;
    private readonly dependencies;
    protected readonly hooks: Map<string, Func[]>;
    protected handleErrors: boolean;
    protected streamCoercionMode: StreamCoercionMode;
    constructor(useGlobal?: boolean);
    addHandlerClass(constructor: ClassConstructor): void;
    /**
     * Iniitalizes all API factories
     *
     * TODO: Move to a helper class
     */
    protected initHandlers(): Map<ApiMethod, Map<string, IApiHandlerInstance<TransportParamsType>>>;
    static GetRouteTokens(route: string): p2r.Token[];
    private getWrappedHandler;
    protected getErrorResponseForException(e: any): IApiInvocationResult;
    protected preProcessInvocationParams(context: IApiInvocationContext<TransportParamsType>, processors: IApiProcessors<TransportParamsType>): Promise<IApiInvocationContext<TransportParamsType>>;
    protected postProcessInvocationResult(context: IApiInvocationContextPostInvoke<TransportParamsType>, processors: IApiProcessors<TransportParamsType>): Promise<IApiInvocationContextPostInvoke<TransportParamsType>>;
    protected isHttpErrorLike(e: any): any;
    getExecutionContext(): IApiInvocationContextPostInvoke<TransportParamsType>;
    protected static getExecutionContext<TransportParamsType extends object>(): IApiInvocationContextPostInvoke<TransportParamsType>;
    static HasExecutionContext(): boolean;
    private invokeHandler;
    private applyValidationFunction;
    protected validateNumberParam({ typedef, name, numberMin, numberMax }: __ApiParamArgs, parsed: any): void;
    protected validateEnumParam({ typedef, name, optional }: __ApiParamArgs, parsed: any): void;
    protected validateRegExpParam(args: __ApiParamArgs, parsed: any): void;
    private getHandlerArg;
    /**
     * Gets a writable stream to pipe output data to
     * @returns a writable stream
     */
    protected getStreamForOutput(def: IApiParamDefinition, invocationParams: IApiInvocationParams<TransportParamsType>): stream.Writable;
    private parseRawRequestBody;
    private validateBodyContents;
    private getApiParam;
    /**
     * Instantiates a class instance of a handler, doing any dependency injection if needed
     * @param handlerClass
     */
    private instanstiateHandlerClass;
    /**
     * Registers a dependency.
     *
     * Dependencies are injected by type. If you'd like to register many dependencies for a type, specify a unique `name` for each dependency.
     *
     * Must be called before initialization to be applied.
     * @param dep The dependency to fill
     * @param instance An instance of the dependency to be used or a function that returns an instance of the dependency.
     */
    addDependency<C extends object = {}>(dep: ClassConstructor<C>, instance?: C | Func0<C>): void;
    /**
     * Gets the value of a header in the current request
     * @param name The name of the header to get. Case insensitive.
     */
    getHeader(name: string): string | string[] | undefined;
    /**
     * Gets the value of a header in the current request
     * @param name The name of the header to get. Case insensitive.
     */
    static getHeader(name: string): string | string[] | undefined;
    /**
     * Sets the value of a header in the current response
     * @param name The name of the header to set. Case insensitive.
     * @param value The value of the header to set
     */
    abstract setHeader(name: string, value: string | number): void;
    /**
     * Sets the HTTP status code for the response
     * @param status The HTTP status code to return in the request
     */
    setResponseStatus(status: number): void;
    /**
     * Sets the HTTP status code for the response
     * @param status The HTTP status code to return in the request
     */
    static setResponseStatus(status: number): void;
    /**
     * Add a hook to be processed during request handling.
     * @param hookName The name of the hook event
     * @param hookfunction The handler to register
     */
    addHook(hookName: ManagedApi<TransportParamsType>['HookHandlerPreInvoke'], hookfunction: ManagedApiPreInvokeHandlerType<TransportParamsType>): void;
    addHook(hookName: ManagedApi<TransportParamsType>['HookHandlerPostInvoke'], hookfunction: ManagedApiPostInvokeHandlerType<TransportParamsType>): void;
    /**
     * Remove a previously registered hook.
     * @param hookName The name of the hook event
     * @param hookfunction The handler to remove
     * @returns true if the hook was found and removed
     */
    removeHook(hookName: string, hookfunction: Func): boolean;
    private hook_checkStreamCoercionMode;
}
