"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ManagedApi = void 0;
const ManagedApiInternal_1 = require("./ManagedApiInternal");
const cls_hooked_1 = require("cls-hooked");
const Errors_1 = require("../Errors");
const HttpError_1 = require("../HttpError");
const MimeTypes_1 = require("./MimeTypes");
const jsonschema_1 = require("jsonschema");
const CallbackPromiseHelper_1 = require("./CallbackPromiseHelper");
const p2r = require("path-to-regexp");
const ApiDependency_1 = require("./ApiDependency");
const ApiProcessing_1 = require("./ApiProcessing/ApiProcessing");
const StreamCoercer_1 = require("../Util/StreamCoercer");
const StreamIntermediary_1 = require("../Util/StreamIntermediary");
const Constants_1 = require("../Constants");
class ManagedApi {
    constructor(useGlobal = false) {
        this.useGlobal = useGlobal;
        this.HookHandlerPreInvoke = 'handler-preinvoke';
        this.HookHandlerPostInvoke = 'handler-postinvoke';
        this.dependencies = new ApiDependency_1.ApiDependencyCollection();
        this.hooks = new Map();
        this.handleErrors = true;
        this.streamCoercionMode = StreamCoercer_1.StreamCoercionMode.Any;
        if (!ManagedApi.namespace) {
            ManagedApi.namespace = (0, cls_hooked_1.getNamespace)(ManagedApi.ClsNamespace);
            if (!ManagedApi.namespace) {
                ManagedApi.namespace = (0, cls_hooked_1.createNamespace)(ManagedApi.ClsNamespace);
            }
        }
        this.classes = [];
        this.jsonValidator = new jsonschema_1.Validator();
        this.addHook('handler-postinvoke', (h) => this.hook_checkStreamCoercionMode(h));
    }
    addHandlerClass(constructor) {
        if (this.useGlobal) {
            throw new Error('addHandlerClass may only be used if ManagedApi is initialized with `useGlobal` = false');
        }
        this.dependencies.registerDependency(ApiDependency_1.ApiDependency.WithConstructor(constructor));
        this.classes.push({
            constructor,
            handlers: ManagedApiInternal_1.ManagedApiInternal.GetHandlersForConstructor(constructor),
        });
    }
    /**
     * Iniitalizes all API factories
     *
     * TODO: Move to a helper class
     */
    initHandlers() {
        if (this.useGlobal) {
            // Get the list of registered classes
            const globalClasses = ManagedApiInternal_1.ManagedApiInternal.GetRegisteredApiClassDefinitions()
                // Only get the ones that have handlers defined on them
                .filter(c => c.handlers.size > 0);
            for (const c of globalClasses) {
                this.dependencies.registerDependency(ApiDependency_1.ApiDependency.WithConstructor(c.constructor));
                this.classes.push(c);
            }
        }
        // Initialize each class and build a master map of all API handlers
        const handlers = new Map();
        for (const handlerClass of this.classes) {
            const instance = this.instanstiateHandlerClass(handlerClass);
            for (const [handlerMethod, handlerMap] of handlerClass.handlers) {
                if (!handlers.has(handlerMethod)) {
                    handlers.set(handlerMethod, new Map());
                }
                const handlerMethodCollection = handlers.get(handlerMethod);
                for (const [route, definition] of handlerMap) {
                    if (handlerMethodCollection.has(route)) {
                        throw ManagedApiInternal_1.ManagedApiInternal.ErrorMultipleApiDefinition(handlerMethod, route);
                    }
                    const routeTokens = ManagedApi.GetRouteTokens(route);
                    const handlerArgs = ManagedApiInternal_1.ManagedApiInternal.GetApiHandlerParams(handlerClass.constructor, definition.handlerKey);
                    handlerMethodCollection.set(route, {
                        parent: instance,
                        wrappedHandler: this.getWrappedHandler({
                            ...definition,
                            processors: ManagedApiInternal_1.ManagedApiInternal.GetApiHandlerProcessors(definition.handler),
                        }, handlerArgs, instance),
                        handlerArgs,
                        routeTokens,
                        ...definition
                    });
                }
            }
        }
        return handlers;
    }
    static GetRouteTokens(route) {
        return p2r.parse(route);
    }
    getWrappedHandler(def, handlerArgs, instance) {
        return (invocationParams) => {
            return ManagedApi.namespace.runPromise(async () => {
                try {
                    const context = {
                        ...await this.preProcessInvocationParams({
                            apiDefinition: def,
                            invocationParams,
                        }, def.processors),
                        result: {
                            statusCode: def.responseCodes?.[0] ?? Constants_1.DefaultApiResponseCode,
                            body: null,
                            headers: {},
                            streamMode: StreamCoercer_1.StreamCoercionMode.None,
                        }
                    };
                    ManagedApi.namespace.set(ManagedApi.ContextNamespace, context);
                    context.result.body = await this.invokeHandler(def, handlerArgs, instance, context.invocationParams);
                    const { result } = await this.postProcessInvocationResult(context, def.processors);
                    return result;
                }
                catch (e) {
                    return this.getErrorResponseForException(e);
                }
                finally {
                    ManagedApi.namespace.set(ManagedApi.ContextNamespace, null);
                }
            });
        };
    }
    getErrorResponseForException(e) {
        if (this.handleErrors) {
            if (this.isHttpErrorLike(e)) {
                return {
                    streamMode: StreamCoercer_1.StreamCoercionMode.None,
                    statusCode: e.statusCode || e.code,
                    body: e.message,
                    headers: {},
                };
            }
            else {
                return {
                    streamMode: StreamCoercer_1.StreamCoercionMode.None,
                    statusCode: 500,
                    body: '',
                    headers: {},
                };
            }
        }
        else {
            throw e;
        }
    }
    async preProcessInvocationParams(context, processors) {
        for (const processor of processors[ApiProcessing_1.ApiProcessorTime.StagePreInvoke]) {
            context.invocationParams = await processor.processor(context.invocationParams);
        }
        if (this.hooks.has(this.HookHandlerPreInvoke)) {
            const handlers = this.hooks.get(this.HookHandlerPreInvoke);
            for (const handler of handlers) {
                const result = await handler(context);
                if (typeof result === 'object') {
                    context = result;
                }
            }
        }
        return context;
    }
    async postProcessInvocationResult(context, processors) {
        for (const processor of processors[ApiProcessing_1.ApiProcessorTime.StagePostInvoke]) {
            context.result = await processor.processor(context.result);
        }
        if (this.hooks.has(this.HookHandlerPostInvoke)) {
            const handlers = this.hooks.get(this.HookHandlerPostInvoke);
            for (const handler of handlers) {
                const result = await handler(context);
                if (typeof result === 'object') {
                    context.result = result;
                }
            }
        }
        return context;
    }
    isHttpErrorLike(e) {
        return e instanceof HttpError_1.HttpError || (typeof e === 'object' && (e.statusCode || e.code) && e.message);
    }
    getExecutionContext() {
        return ManagedApi.namespace?.get(ManagedApi.ContextNamespace);
    }
    static getExecutionContext() {
        return ManagedApi.namespace?.get(ManagedApi.ContextNamespace);
    }
    static HasExecutionContext() {
        return Boolean(ManagedApi.getExecutionContext());
    }
    async invokeHandler(def, handlerArgs, instance, invocationParams) {
        // Resolve handler arguments
        let useCallback;
        let output = null, overrideOutput = false;
        const args = await Promise.all(handlerArgs.map(async (def) => {
            // Callback arg is special
            if (def.type === 5 /* ApiParamType.Callback */) {
                if (useCallback) {
                    throw new Error('Only a single callback parameter may be defined on a handler');
                }
                useCallback = new CallbackPromiseHelper_1.PromiseCallbackHelper();
                return await useCallback.getCallbackFunc();
            }
            // Parse the contents of the argument and validate
            const parsed = await this.getHandlerArg(def, invocationParams);
            if (def.type === 8 /* ApiParamType.Out */ && def.overrideOutput) {
                if (overrideOutput) {
                    throw new Error('Cannot override output in multiple args');
                }
                overrideOutput = true;
                output = parsed;
            }
            if (!def.args || (typeof parsed === 'undefined' && def.args.optional)) {
                return parsed;
            }
            this.validateEnumParam(def.args, parsed);
            if (def.args.validationFunc) {
                this.applyValidationFunction(def.args, parsed);
            }
            if (def.args.regexp) {
                this.validateRegExpParam(def.args, parsed);
            }
            if (def.args.typedef?.type === 'number') {
                this.validateNumberParam(def.args, parsed);
            }
            if (def.args.validationFunc) {
                try {
                    def.args.validationFunc(def.args.name, parsed);
                }
                catch (e) {
                    if (this.isHttpErrorLike(e)) {
                        throw e;
                    }
                    else {
                        throw new Errors_1.HttpParamInvalidError(def.args.name);
                    }
                }
            }
            return parsed;
        }));
        let executionResult;
        if (useCallback) {
            executionResult = await useCallback.execute(() => def.handler.apply(instance, args));
        }
        else {
            executionResult = await def.handler.apply(instance, args);
        }
        if (overrideOutput) {
            executionResult = await output;
        }
        return executionResult;
    }
    applyValidationFunction({ name, validationFunc }, parsed) {
        if (validationFunc) {
            validationFunc(name, parsed);
        }
    }
    validateNumberParam({ typedef, name, numberMin, numberMax }, parsed) {
        if (typedef?.type === 'number') {
            if ((typeof numberMin === 'number' && parsed < numberMin)
                || (typeof numberMax === 'number' && parsed > numberMax)) {
                throw new Errors_1.HttpNumberParamOutOfBoundsError(name, numberMin, numberMax);
            }
        }
    }
    validateEnumParam({ typedef, name, optional }, parsed) {
        if (typeof parsed === 'undefined' && optional) {
            return;
        }
        if (typedef?.type === 'number' || typedef?.type === 'string' || typedef?.type === 'enum') {
            if (typedef.schema) {
                if ('enum' in typedef.schema && typedef.schema.enum.indexOf(parsed) === -1) {
                    throw new Errors_1.HttpEnumParamInvalidValueError(name, typedef.schema.enum);
                }
                if ('const' in typedef.schema && typeof typedef.schema.const !== 'undefined' && typedef.schema.const !== parsed) {
                    throw new Errors_1.HttpConstParamInvalidValueError(name, typedef.schema.const);
                }
            }
        }
    }
    validateRegExpParam(args, parsed) {
        if (args.regexp) {
            let regexp;
            if (typeof args.regexp === 'function') {
                regexp = args.regexp(args.name);
            }
            else {
                regexp = args.regexp;
            }
            if (!regexp.test(parsed)) {
                throw new Errors_1.HttpRegexParamInvalidTypeError(args.name, regexp.source);
            }
        }
    }
    async getHandlerArg(def, invocationParams) {
        const { args } = def;
        if (def.type === 2 /* ApiParamType.Query */) {
            const isDefined = typeof invocationParams.queryParams[args.name] === 'string';
            if (!isDefined && args.typedef.type !== 'boolean') {
                // Query param not defined
                if (args.initializer) {
                    return args.initializer();
                }
                else if (args.optional) {
                    // This arg was optional
                    return undefined;
                }
                else {
                    // Arg was required, throw
                    throw new Errors_1.HttpRequiredQueryParamMissingError(args.name);
                }
            }
            // Arg was defined, process it
            return this.getApiParam(args, isDefined, invocationParams.queryParams[args.name]);
        }
        else if (def.type === 3 /* ApiParamType.Path */) {
            const isDefined = typeof invocationParams.pathParams[args.name] === 'string';
            if (!isDefined && args.typedef.type !== 'boolean') {
                // Path param not defined
                if (args.initializer) {
                    return args.initializer();
                }
                else if (args.optional) {
                    // This arg was optional
                    return undefined;
                }
                else {
                    // Arg was required, throw
                    throw new Errors_1.HttpRequiredPathParamMissingError(args.name);
                }
            }
            // Arg was defined, process it
            return this.getApiParam(args, isDefined, invocationParams.pathParams[args.name]);
        }
        else if (def.type === 4 /* ApiParamType.Header */) {
            const headerName = args.name.toLowerCase();
            const isDefined = typeof invocationParams.headers[headerName] === 'string';
            if (!isDefined && args.typedef.type !== 'boolean') {
                // Query param not defined
                if (args.initializer) {
                    return args.initializer();
                }
                else if (args.optional) {
                    // This arg was optional
                    return undefined;
                }
                else {
                    // Arg was required, throw
                    throw new Errors_1.HttpRequiredHeaderParamMissingError(args.name);
                }
            }
            // Arg was defined, process it
            return this.getApiParam(args, isDefined, invocationParams.headers[headerName]);
        }
        else if (def.type === 0 /* ApiParamType.Body */ || def.type === 1 /* ApiParamType.RawBody */) {
            const { bodyContents } = invocationParams;
            if (!bodyContents) {
                if (args.initializer) {
                    return args.initializer();
                }
                throw new Errors_1.HttpRequiredBodyParamMissingError();
            }
            if (def.type === 1 /* ApiParamType.RawBody */) {
                if (typeof def.mimeType === 'string' && def.mimeType !== bodyContents.streamContentsMimeRaw) {
                    throw new Errors_1.HttpUnsupportedMediaTypeError(def.mimeType);
                }
                if (def.bodyType === 0 /* ApiRawBodyParamType.Stream */) {
                    if (!bodyContents.contentsStream) {
                        throw new Errors_1.HttpTransportConfigurationError('Raw body was not available for handler');
                    }
                    return bodyContents.contentsStream;
                }
                else if (def.bodyType === 1 /* ApiRawBodyParamType.String */) {
                    if (!bodyContents.contentsString) {
                        bodyContents.contentsString = await bodyContents.readStreamToStringAsync();
                    }
                    return bodyContents.contentsString;
                }
                else {
                    throw new Errors_1.HttpTransportConfigurationError('Unknown raw body type requested');
                }
            }
            let contents = null;
            if (typeof bodyContents.parsedBody !== 'undefined') {
                contents = bodyContents.parsedBody;
                if (contents instanceof Buffer) {
                    throw new Error('ParsedBody may not be a buffer');
                }
            }
            else {
                contents = await this.parseRawRequestBody(bodyContents);
                switch (args.typedef.type) {
                    case 'any':
                        if (typeof contents === 'object') {
                            break;
                        }
                    case 'string':
                    case 'number':
                    case 'boolean':
                        contents = this.getApiParam(args, true, contents);
                        break;
                    case 'object':
                    case 'array':
                        break;
                    default:
                        throw new Error('Not implemented');
                }
            }
            this.validateBodyContents(args, contents);
            return contents;
        }
        else if (def.type === 6 /* ApiParamType.Transport */) {
            if (invocationParams.transportParams[def.transportTypeId]) {
                return invocationParams.transportParams[def.transportTypeId];
            }
            else if (args?.initializer) {
                return args.initializer();
            }
            else if (!args?.optional) {
                throw new Errors_1.HttpRequiredTransportParamMissingError(def.transportTypeId);
            }
            else {
                return undefined;
            }
        }
        else if (def.type === 7 /* ApiParamType.Dependency */) {
            try {
                if (args.typeref) {
                    return this.dependencies.instantiateDependency(args.typeref);
                }
                else {
                    return this.dependencies.instantiateDependency(args.typedef);
                }
            }
            catch (e) {
                if (args.initializer) {
                    return args.initializer();
                }
                else if (args.optional) {
                    return undefined;
                }
                throw e;
            }
        }
        else if (def.type === 8 /* ApiParamType.Out */) {
            if (this.streamCoercionMode === StreamCoercer_1.StreamCoercionMode.None) {
                throw new Error('Stream out type cannot be used while coercion is disabled');
            }
            return this.getStreamForOutput(def, invocationParams);
        }
    }
    /**
     * Gets a writable stream to pipe output data to
     * @returns a writable stream
     */
    getStreamForOutput(def, invocationParams) {
        return StreamIntermediary_1.StreamIntermediary.GetStreamIntermediary(this);
    }
    async parseRawRequestBody(bodyContents) {
        if (!bodyContents.contentsString) {
            bodyContents.contentsString = await bodyContents.readStreamToStringAsync();
        }
        switch (bodyContents.streamContentsMimeType) {
            case MimeTypes_1.ApiMimeType.ApplicationJson:
                return JSON.parse(bodyContents.contentsString);
            case MimeTypes_1.ApiMimeType.Text:
                return bodyContents.contentsString;
            // TODO: Add a way to register mime type parsers so that you can BYO xml
            default:
                throw new Errors_1.HttpUnsupportedMediaTypeError();
        }
    }
    validateBodyContents(args, contents) {
        const contentsType = typeof contents;
        if (args.typedef.type !== 'any' && contentsType !== args.typedef.type) {
            throw new Errors_1.HttpBodyParamInvalidTypeError(args.typedef.type);
        }
        if (args.typedef.type === 'object') {
            const result = this.jsonValidator.validate(contents, args.typedef.schema);
            if (!result.valid) {
                if (result.errors.length === 0) {
                    throw new Errors_1.HttpBodyParamInvalidTypeError(args.typedef.type);
                }
                else {
                    // TODO: Create a multiple error format
                    throw new Errors_1.HttpBodyParamValidationError(result.errors[0]);
                }
            }
        }
        if (args.validationFunc) {
            args.validationFunc(args.name, contents);
        }
    }
    getApiParam(args, isDefined, strVal, contentsMime) {
        switch (args.typedef.type) {
            case 'any':
            case 'string':
                {
                    if (typeof strVal === 'string') {
                        return strVal;
                    }
                    else if (Array.isArray(strVal)) {
                        return strVal[0];
                    }
                    else if (strVal instanceof Buffer) {
                        return strVal.toString();
                    }
                    throw new Errors_1.HttpQueryParamInvalidTypeError(args.name, 'string');
                }
            case 'number':
                {
                    if (Array.isArray(strVal)) {
                        strVal = strVal[0];
                    }
                    else if (strVal instanceof Buffer) {
                        strVal = strVal.toString();
                    }
                    const parsed = Number(strVal);
                    if (isNaN(parsed)) {
                        throw new Errors_1.HttpQueryParamInvalidTypeError(args.name, 'number');
                    }
                    return parsed;
                }
                ;
            case 'boolean':
                {
                    if (isDefined) {
                        switch ((typeof strVal === 'string' ? strVal : strVal.toString()).toLowerCase()) {
                            case '0':
                            case 'false':
                                return false;
                        }
                        return true;
                    }
                    else if (args.initializer) {
                        return args.initializer();
                    }
                    return false;
                }
                ;
            case 'array':
                {
                    // TODO: Check the type of the array elements
                    if (typeof strVal === 'string') {
                        return [strVal];
                    }
                    else if (Array.isArray(strVal)) {
                        return strVal;
                    }
                    else if (strVal instanceof Buffer) {
                        return [strVal.toString()];
                    }
                    throw new Errors_1.HttpQueryParamInvalidTypeError(args.name, 'array');
                }
            default:
                throw new Error('Not implemented');
        }
    }
    /**
     * Instantiates a class instance of a handler, doing any dependency injection if needed
     * @param handlerClass
     */
    instanstiateHandlerClass(handlerClass) {
        const instance = this.dependencies.instantiateDependency(handlerClass.constructor);
        return instance;
    }
    /**
     * Registers a dependency.
     *
     * Dependencies are injected by type. If you'd like to register many dependencies for a type, specify a unique `name` for each dependency.
     *
     * Must be called before initialization to be applied.
     * @param dep The dependency to fill
     * @param instance An instance of the dependency to be used or a function that returns an instance of the dependency.
     */
    addDependency(dep, instance) {
        // TODO: Dependency scopes
        if (instance) {
            this.dependencies.registerDependency(ApiDependency_1.ApiDependency.WithFunc(dep, typeof instance === 'function' ? instance : () => instance));
        }
        else {
            this.dependencies.registerDependency(ApiDependency_1.ApiDependency.WithConstructor(dep));
        }
    }
    /**
     * Gets the value of a header in the current request
     * @param name The name of the header to get. Case insensitive.
     */
    getHeader(name) {
        const context = this.getExecutionContext();
        const headers = context.invocationParams.headers;
        if (!headers) {
            return;
        }
        return headers[name];
    }
    /**
     * Gets the value of a header in the current request
     * @param name The name of the header to get. Case insensitive.
     */
    static getHeader(name) {
        const context = this.getExecutionContext();
        const headers = context.invocationParams.headers;
        if (!headers) {
            return;
        }
        return headers[name];
    }
    /**
     * Sets the HTTP status code for the response
     * @param status The HTTP status code to return in the request
     */
    setResponseStatus(status) {
        const context = this.getExecutionContext();
        context.result.statusCode = status;
    }
    /**
     * Sets the HTTP status code for the response
     * @param status The HTTP status code to return in the request
     */
    static setResponseStatus(status) {
        const context = this.getExecutionContext();
        context.result.statusCode = status;
    }
    addHook(hookName, hookfunction) {
        switch (hookName) {
            case this.HookHandlerPreInvoke:
            case this.HookHandlerPostInvoke:
                if (!this.hooks.has(hookName)) {
                    this.hooks.set(hookName, []);
                }
                this.hooks.get(hookName).push(hookfunction);
                break;
            default:
                throw new Error(`Invalid hook name: ${hookName}`);
        }
    }
    /**
     * Remove a previously registered hook.
     * @param hookName The name of the hook event
     * @param hookfunction The handler to remove
     * @returns true if the hook was found and removed
     */
    removeHook(hookName, hookfunction) {
        if (this.hooks.has(hookName)) {
            const hooks = this.hooks.get(hookName);
            const index = hooks.indexOf(hookfunction);
            if (index >= 0) {
                hooks.splice(index, 1);
                if (hooks.length === 0) {
                    this.hooks.delete(hookName);
                }
                return true;
            }
        }
        return false;
    }
    hook_checkStreamCoercionMode(pi) {
        if (this.streamCoercionMode === StreamCoercer_1.StreamCoercionMode.None) {
            return;
        }
        pi.result.streamMode = StreamCoercer_1.StreamCoercer.IsCoercible(pi.result.body, this.streamCoercionMode);
    }
}
exports.ManagedApi = ManagedApi;
ManagedApi.ClsNamespace = 'ManagedApiNamespace';
ManagedApi.ContextNamespace = 'context';
ManagedApi.apis = [];
//# sourceMappingURL=ManagedApi.js.map