"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ManagedApiInternal = void 0;
require("reflect-metadata");
const ApiProcessing_1 = require("./ApiProcessing/ApiProcessing");
const SINGLETON_KEY = Symbol.for("MB.ts-api-decorators.ManagedApiInternal");
// check if the global object has this symbol
// add it if it does not have the symbol, yet
// ------------------------------------------
const globalSymbols = Object.getOwnPropertySymbols(global);
const keyDefined = (globalSymbols.indexOf(SINGLETON_KEY) > -1);
if (!keyDefined) {
    global[SINGLETON_KEY] = {
        apis: new Map(),
    };
}
let Apis = global[SINGLETON_KEY];
class ManagedApiInternal {
    static ResetRegisteredApis() {
        Apis.apis.clear();
    }
    static RegisterApi(constructor) {
        if (Apis.apis.has(constructor)) {
            throw new Error('Api already registered');
        }
        Apis.apis.set(constructor, {
            constructor,
            handlers: new Map(),
        });
        Reflect.defineMetadata(ManagedApiInternal.ApiMetadataKey, constructor.name, constructor);
    }
    static GetRegisteredApiClassDefinitions() {
        const result = [];
        for (const api of Apis.apis) {
            const handlers = this.GetHandlersForConstructor(api[0]);
            if (handlers) {
                api[1].handlers = handlers;
                result.push(api[1]);
            }
        }
        return Array.from(Apis.apis.values());
    }
    static GetHandlersForConstructor(constructor) {
        const apis = ManagedApiInternal.GetApiDefinitionsOnObject(constructor);
        const handlers = new Map();
        if (!apis || apis.length === 0) {
            return handlers;
        }
        for (const api of apis) {
            if (!handlers.has(api.method)) {
                handlers.set(api.method, new Map());
            }
            const methodHandler = handlers.get(api.method);
            if (methodHandler.has(api.route)) {
                throw ManagedApiInternal.ErrorMultipleApiDefinition(api.method, api.route);
            }
            methodHandler.set(api.route, api);
        }
        return handlers;
    }
    static GetDependenciesOnConstructor(constructor) {
        return this.GetDependencyDefinitionsOnObject(constructor);
    }
    static ErrorMultipleApiDefinition(method, route) {
        return new Error(`Multiple APIs defined for '${method}' method on '${route}'`);
    }
    static AddApiMetadataToObject(metadata, target) {
        const apis = ManagedApiInternal.GetApiDefinitionsOnObject(target);
        apis.push(metadata);
        Reflect.defineMetadata(this.ApiMethodMetadataKey, apis, target);
    }
    static AddApiModifierMetadataToObject(metadata, target) {
        const apis = ManagedApiInternal.GetApiModifierDefinitionsOnObject(target, metadata.propertyKey);
        apis.push(metadata);
        Reflect.defineMetadata(this.ApiMethodModifierMetadataKey, apis, target, metadata.propertyKey);
    }
    static AddGlobalApiProcessorMetadataToObject(metadata, target) {
        const apis = ManagedApiInternal.GetGlobalApiProcessorOnObject(target);
        apis.push(metadata);
        Reflect.defineMetadata(this.GlobalApiProcessorMetadataKey, apis, target);
    }
    static AddApiProcessorsToObject(processors, target) {
        const apis = ManagedApiInternal.GetApiProcessorsOnObject(target);
        apis.push(...processors);
        Reflect.defineMetadata(this.ApiMethodProcessorsMetadataKey, apis, target);
    }
    static AddApiHandlerParamMetadataToObject(metadata, target) {
        const apis = ManagedApiInternal.GetHandlerParamDefinitionsOnObject(target, metadata.propertyKey);
        apis.push(metadata);
        Reflect.defineMetadata(this.ApiMethodParamsMetadataKey, apis, target, metadata.propertyKey);
    }
    static AddApiDependencyParamMetadataToObject(metadata, target) {
        const deps = ManagedApiInternal.GetDependencyParamDefinitionsOnObject(target, metadata.propertyKey);
        deps.push(metadata);
        Reflect.defineMetadata(this.DependencyParamMetadataKey, deps, target, metadata.propertyKey);
    }
    static AddDependencyMetadataToObject(metadata, target) {
        const deps = ManagedApiInternal.GetDependencyDefinitionsOnObject(target);
        deps.push(metadata);
        Reflect.defineMetadata(this.DependencyMetadataKey, deps, target);
    }
    static GetDependencyParams(target, key) {
        const params = this.GetDependencyParamDefinitionsOnObject(target, key)
            .sort((a, b) => a.parameterIndex - b.parameterIndex);
        for (let i = 0; i < params.length; ++i) {
            if (params[i].parameterIndex !== i) {
                throw new Error(`All parameters must be decorated for handler: ${String(key)}`);
            }
        }
        return params;
    }
    static GetApiHandlerParams(target, key) {
        const params = this.GetHandlerParamDefinitionsOnObject(target, key)
            .sort((a, b) => a.parameterIndex - b.parameterIndex);
        for (let i = 0; i < params.length; ++i) {
            if (params[i].parameterIndex !== i) {
                throw new Error(`All parameters must be decorated for handler: ${String(key)}`);
            }
        }
        return params;
    }
    static GetApiHandlerProcessors(target) {
        const processors = {
            [ApiProcessing_1.ApiProcessorTime.StagePreInvoke]: [],
            [ApiProcessing_1.ApiProcessorTime.StagePostInvoke]: [],
        };
        for (const processor of ManagedApiInternal.GetApiProcessorsOnObject(target)) {
            processors[processor.stage].push(processor);
        }
        return processors;
    }
    static GetApiDefinitionsOnObject(target) {
        return Reflect.getMetadata(this.ApiMethodMetadataKey, target) || [];
    }
    static GetApiModifierDefinitionsOnObject(target, propertyKey) {
        return Reflect.getMetadata(this.ApiMethodModifierMetadataKey, target, propertyKey) || [];
    }
    static GetGlobalApiProcessorOnObject(target) {
        return Reflect.getMetadata(this.GlobalApiProcessorMetadataKey, target) || [];
    }
    static GetApiProcessorsOnObject(target) {
        return Reflect.getMetadata(this.ApiMethodProcessorsMetadataKey, target) || [];
    }
    static GetHandlerParamDefinitionsOnObject(target, key) {
        return Reflect.getMetadata(this.ApiMethodParamsMetadataKey, target, key) || [];
    }
    static GetDependencyDefinitionsOnObject(target) {
        return Reflect.getMetadata(this.DependencyMetadataKey, target) || [];
    }
    static GetDependencyParamDefinitionsOnObject(target, key) {
        return Reflect.getMetadata(this.DependencyParamMetadataKey, target, key) || [];
    }
}
exports.ManagedApiInternal = ManagedApiInternal;
ManagedApiInternal.ApiMetadataKey = 'managedapi:api';
ManagedApiInternal.ApiMethodMetadataKey = 'managedapi:apimethod';
ManagedApiInternal.ApiMethodModifierMetadataKey = 'managedapi:apimodifier';
ManagedApiInternal.ApiMethodProcessorsMetadataKey = 'managedapi:apiprocessors';
ManagedApiInternal.ApiMethodParamsMetadataKey = 'managedapi:apimethodparams';
ManagedApiInternal.DependencyMetadataKey = 'managedapi:dependency';
ManagedApiInternal.DependencyParamMetadataKey = 'managedapi:dependencyparams';
ManagedApiInternal.GlobalApiProcessorMetadataKey = 'managedapi:globalapiprocessor';
//# sourceMappingURL=ManagedApiInternal.js.map