"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ExtractCommand = void 0;
const Swagger2Extractor_1 = require("./Swagger2Extractor");
const CliCommand_1 = require("./CliCommand");
const OpenApiV3_1 = require("./OpenApiV3");
const ApiParser_1 = require("../Util/ApiParser");
class ExtractCommand extends CliCommand_1.CliCommand {
    constructor(program) {
        super();
        this.program = program;
        this.apiParser = new ApiParser_1.ApiParser();
        program
            .command('extract <rootDir>')
            .description('Extract api information from the program at rootDir.', {
            'rootDir': 'The root directory or file of the program to extract. If directory, should be the directory that contains tsConfig',
        })
            .option('--tsconfig <file>', 'The tsconfig.json file to use when compiling')
            .option('--type <type>', 'The type of output to generate', (d, v) => this.validateProgramOutputType(d, v), 'openapiv3')
            .option('--outFile', 'The file to write output to')
            .option('--silent', 'Don\'t output information', false)
            .option('--apiInfo <file>', 'File containing information for the API', 'package.json')
            .action((rootDir, options) => this.runCommand({
            ...options,
            rootDir,
        }));
    }
    async runCommand(options) {
        const api = await this.apiParser.parseApi(options);
        let summary;
        switch (options.type) {
            case 'summary':
                summary = this.printExtractionSummary(options, api, this.apiParser.getTree());
                break;
            case 'swagger2':
                summary = this.getSwaggerSummary(options, api);
                break;
            case 'openapiv3':
                summary = this.getOpenApiV3Summary(options, api);
                break;
            case 'json':
                summary = this.printJsonSummary(options, api);
                break;
            default:
                throw new Error('Unknown output type: ' + options.type);
        }
        this.printSummary(summary);
    }
    printSummary(summary) {
        console.log(summary instanceof Buffer ? summary.toString('utf8') : summary);
    }
    printJsonSummary(options, api) {
        throw new Error("Method not implemented.");
    }
    getSwaggerSummary(options, api) {
        const extractor = new Swagger2Extractor_1.Swagger2Extractor(api.tree, api.programInfo, {});
        return extractor.toString();
    }
    getOpenApiV3Summary(options, api) {
        const extractor = new OpenApiV3_1.OpenApiV3Extractor(api.tree, api.programInfo, {});
        return extractor.toString();
    }
    validateProgramOutputType(value, defaultValue) {
        switch (value) {
            case 'summary':
            case 'json':
            case 'swagger2':
            case 'openapiv3':
                return value;
            case undefined:
                return defaultValue;
            default:
                throw new Error('Invalid value for `type`: ' + value);
        }
    }
}
exports.ExtractCommand = ExtractCommand;
//# sourceMappingURL=ExtractCommand.js.map