"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OpenApiV3JsonExtractor = exports.OpenApiV3Extractor = void 0;
const TransformerMetadata_1 = require("../transformer/TransformerMetadata");
const yaml = require("js-yaml");
const HandlerTree_1 = require("../transformer/HandlerTree");
const apiManagement_1 = require("../apiManagement");
class OpenApiV3Extractor {
    constructor(apiTree, apiInfo, opts) {
        this.apiTree = apiTree;
        this.apiInfo = apiInfo;
        this.opts = opts;
        this.removableTypes = ['undefined', 'null'];
        this.tags = new Map();
        this.definitions = new Map();
        this.definitionNameMap = new Map();
        this.expectedDefinitions = new Set();
    }
    getDocument() {
        const doc = {
            openapi: OpenApiV3Extractor.SwaggerVersion,
            info: {
                title: this.apiInfo.title,
                version: this.apiInfo.version,
                description: this.apiInfo.description,
                license: this.apiInfo.license ? this.apiInfo.license[0] : undefined,
                termsOfService: this.apiInfo.termsOfService,
                contact: this.apiInfo.contact,
            },
            servers: this.getApiServers(),
            paths: this.getPaths(),
        };
        if (this.tags.size > 0) {
            doc.tags = [];
            this.tags.forEach(tag => doc.tags.push(this.getTagObject(tag)));
        }
        if (this.definitions.size > 0) {
            doc.components = this.getComponents();
        }
        return doc;
    }
    getApiServers() {
        const schemes = this.apiInfo.schemes || this.tryInferSchemes() || [];
        if (schemes.length === 0) {
            schemes.push('');
        }
        let host = this.validHost(this.apiInfo.host || this.apiInfo.homepage) || '';
        while (host.endsWith('/')) {
            host = host.substr(0, host.length - 1);
        }
        let basePath = this.apiInfo.basePath;
        if (basePath && !basePath.startsWith('/')) {
            basePath = '/' + basePath;
        }
        const servers = [];
        for (const scheme of schemes) {
            let url = host + this.apiInfo.basePath;
            if (scheme.length > 0) {
                url = `${scheme}://${url}`;
            }
            servers.push({
                url: url,
            });
        }
        return servers;
    }
    getComponents() {
        const defKeys = Array.from(this.definitions.keys());
        defKeys.sort();
        const definitions = {
            schemas: {},
        };
        defKeys.forEach(defName => {
            definitions.schemas[defName] = this.definitions.get(defName);
        });
        return definitions;
    }
    tryInferSchemes() {
        if (this.opts.disableTryInferSchemes) {
            return;
        }
        const schemes = new Set();
        if (this.apiInfo.host) {
            const url = new URL(this.apiInfo.host);
            if (url.protocol.length > 0 && !schemes.has(url.protocol)) {
                schemes.add(url.protocol.replace(':', ''));
            }
        }
        if (schemes.size > 0) {
            return Array.from(schemes);
        }
    }
    validHost(host) {
        try {
            const url = new URL(host);
            if (url.host.length > 0) {
                return url.host;
            }
        }
        catch (e) {
        }
        return host;
    }
    getPaths() {
        const paths = {};
        // Array.from(WalkChildrenByType(api, isHandlerParameterNode)).forEach(api => {
        for (const api of (0, HandlerTree_1.WalkTreeByType)(this.apiTree, HandlerTree_1.isHandlerNode)) {
            const operation = this.getOperationObject(api);
            if (!operation) {
                continue;
            }
            const route = this.swaggerizeRoute(api.route);
            if (!paths[route]) {
                paths[route] = {};
            }
            if (!paths[route][api.apiMethod.toLowerCase()]) {
                paths[route][api.apiMethod.toLowerCase()] = operation;
            }
            else {
                throw new Error(`Multiple APIs for route: [${api.apiMethod}]: ${route}`);
            }
        }
        ;
        return paths;
    }
    swaggerizeRoute(route) {
        return apiManagement_1.ManagedApi.GetRouteTokens(route)
            .map(t => {
            if (typeof t === 'string') {
                return t;
            }
            let outStr = t.prefix ? t.prefix : '';
            if (t.modifier && t.modifier !== '?') {
                console.warn('Unhandled modifier: ' + t.modifier);
            }
            outStr += `{${t.name}${t.modifier ? t.modifier : ''}}`;
            if (t.suffix) {
                outStr += t.suffix;
            }
            return outStr;
        })
            .join('');
    }
    getOperationObject(api) {
        if ((0, TransformerMetadata_1.getMetadataValue)(api.metadata, "OpenApi" /* IMetadataType.OpenApi */, undefined, "private" /* OpenApiMetadataType.Private */)) {
            return;
        }
        const params = Array.from((0, HandlerTree_1.WalkChildrenByType)(api, HandlerTree_1.isHandlerParameterNode));
        const bodyParams = params.filter(p => this.isBodyParam(p));
        const metadataTags = (0, TransformerMetadata_1.getAllMetadataValues)(api.metadata, "OpenApi" /* IMetadataType.OpenApi */, undefined, "tag" /* OpenApiMetadataType.Tag */);
        const response = this.getResponseObject(api);
        const responses = {};
        let responseCodes = (0, TransformerMetadata_1.getMetadataValueByDescriptor)(api.metadata, TransformerMetadata_1.BuiltinMetadata.ResponseCodes);
        if (!Array.isArray(responseCodes) || responseCodes.length === 0) {
            responseCodes = ['default'];
        }
        for (const responseCode of responseCodes) {
            responses[responseCode] = response;
        }
        return {
            operationId: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(api.metadata, TransformerMetadata_1.BuiltinMetadata.Name),
            description: (0, TransformerMetadata_1.getMetadataValue)(api.metadata, "OpenApi" /* IMetadataType.OpenApi */, undefined, "description" /* OpenApiMetadataType.Description */),
            summary: (0, TransformerMetadata_1.getMetadataValue)(api.metadata, "OpenApi" /* IMetadataType.OpenApi */, undefined, "summary" /* OpenApiMetadataType.Summary */),
            tags: metadataTags.map(t => this.recordTagObject(t)),
            parameters: params
                .filter(p => !this.isBodyParam(p))
                .map(p => this.getParametersObject(p))
                .filter(p => !!p),
            requestBody: bodyParams ? this.getRequestBody(bodyParams) : undefined,
            responses,
        };
    }
    isBodyParam(p) {
        return p.paramDef.type === 0 /* ApiParamType.Body */ || p.paramDef.type === 1 /* ApiParamType.RawBody */ || p.paramDef.type === 10 /* ApiParamType.FormFileSingle */;
    }
    getResponseObject(api) {
        const schema = (0, TransformerMetadata_1.getMetadataValueByDescriptor)(api.metadata, TransformerMetadata_1.BuiltinMetadata.ReturnSchema);
        let mimeType = OpenApiV3Extractor.MimeTypeText;
        switch (schema.type) {
            case 'object':
            case 'array':
            case 'intersection':
            case 'union':
                mimeType = OpenApiV3Extractor.MimeTypeJson;
                break;
        }
        const response = {
            description: this.getOperationResponseDescription(api, schema),
            content: {
                [mimeType]: {
                    schema: this.getInlineTypeSchema(schema),
                }
            }
        };
        return response;
    }
    getOperationResponseDescription(api, schema) {
        const metadata = (0, TransformerMetadata_1.getMetadataValue)(api.metadata, "OpenApi" /* IMetadataType.OpenApi */, undefined, "response" /* OpenApiMetadataType.ResponseDescription */);
        if (metadata) {
            return metadata;
        }
        if (schema) {
            if (schema.typename) {
                return schema.typename;
            }
            return `${schema.type} response`;
        }
        return 'Default response';
    }
    getInlineTypeSchema(returnType) {
        switch (returnType.type) {
            case null:
            // @ts-ignore
            case 'null':
                return {
                    type: null,
                };
            case 'array':
                if (returnType.elementType) {
                    return {
                        type: 'array',
                        items: this.getInlineTypeSchema(returnType.elementType),
                    };
                }
                else {
                    throw new Error(`Internal Error: Array types must have an elementType defined: ${returnType.typename}`);
                }
                break;
            case 'external':
                return returnType.schema;
            case 'object':
                if (returnType.schema) {
                    return this.getInternalSchemaToOutputSchema(returnType.schema);
                }
            // Fall through to next case is intentional
            case 'number':
            case 'string':
                return {
                    type: returnType.type,
                    enum: this.getReturnTypeEnumValue(returnType),
                };
            case 'enum':
                return this.getEnumTypeInline(returnType);
            case 'boolean':
                return {
                    type: returnType.type,
                };
            case 'void':
                return {};
            default:
                throw new Error(`Unable to serialize inline type: ${returnType.typename} (type: ${returnType.type})`);
        }
    }
    getReturnTypeEnumValue(returnType) {
        if (returnType.schema) {
            if ('enum' in returnType.schema) {
                return returnType.schema.enum;
            }
            if ('const' in returnType.schema) {
                return [returnType.schema.const];
            }
        }
        return undefined;
    }
    getEnumTypeInline(returnType) {
        if (!returnType.schema || !returnType.schema.enum || returnType.schema.enum.length === 0) {
            // TODO: this has issues for named types and otehr scenarios
            return {
                type: 'string',
            };
        }
        if (returnType.typename) {
            this.addDefinitions({
                [returnType.typename]: {
                    enum: returnType.schema.enum,
                }
            });
            return {
                $ref: `#/components/schemas/abcd___${this.renameDefinition(returnType.typename)}`,
            };
        }
        const types = new Map();
        for (const val of returnType.schema.enum) {
            const type = typeof val;
            if (!types.has(type)) {
                types.set(type, []);
            }
            types.get(type).push(val);
        }
        if (types.size === 1) {
            return {
                type: typeof returnType.schema.enum[0],
                enum: returnType.schema.enum,
            };
        }
        const enumType = {
            type: 'string',
            anyOf: [],
        };
        types.forEach((values, type) => {
            enumType.anyOf.push({
                type: type,
                enum: values,
            });
        });
        return enumType;
    }
    getInternalSchemaToOutputSchema(schema) {
        if (typeof schema.definitions !== 'boolean' && schema.definitions) {
            this.addDefinitions(schema.definitions);
            schema = { ...schema };
            delete schema.definitions;
        }
        const delProps = ['uniqueTypename', 'typename', '$schema'];
        for (const prop of delProps) {
            if (typeof schema[prop] !== 'undefined') {
                delete schema[prop];
            }
        }
        if (schema.$ref) {
            schema.$ref = this.replaceRefStr(schema.$ref);
        }
        return schema;
    }
    addDefinitions(definitions) {
        for (const definition of Object.keys(definitions)) {
            const renamedDef = this.renameDefinition(definition);
            this.definitions.set(renamedDef, this.fixDefinitionProperties(this.fixDefinitionRefs(definitions[definition])));
        }
    }
    renameDefinition(definition) {
        if (!this.definitionNameMap.has(definition)) {
            let defName = definition.includes('.') ? definition.substring(0, definition.lastIndexOf('.')) : definition;
            if (definition.includes('.')) {
                let i = 0;
                while (this.definitions.has(defName)) {
                    defName = `${defName}${++i}`;
                }
            }
            this.definitionNameMap.set(definition, defName);
        }
        return this.definitionNameMap.get(definition);
    }
    fixDefinitionProperties(_) {
        const redef = _;
        if (redef.anyOf) {
            redef.anyOf = this._fixDefinitionPropertiesCollection(redef.anyOf);
        }
        if (redef.allOf) {
            redef.allOf = this._fixDefinitionPropertiesCollection(redef.allOf);
        }
        if (redef.oneOf) {
            redef.oneOf = this._fixDefinitionPropertiesCollection(redef.oneOf);
        }
        if (redef.properties) {
            redef.properties = this._fixDefinitionPropertiesCollection(redef.properties);
        }
        if (redef.definitions) {
            redef.definitions = this._fixDefinitionPropertiesCollection(redef.definitions);
        }
        if (typeof redef.additionalProperties === 'object') {
            if (redef.additionalProperties.anyOf) {
                redef.additionalProperties.anyOf = this._fixDefinitionPropertiesCollection(redef.additionalProperties.anyOf);
            }
            if (redef.additionalProperties.allOf) {
                redef.additionalProperties.allOf = this._fixDefinitionPropertiesCollection(redef.additionalProperties.allOf);
            }
            if (redef.additionalProperties.oneOf) {
                redef.additionalProperties.oneOf = this._fixDefinitionPropertiesCollection(redef.additionalProperties.oneOf);
            }
            if (redef.additionalProperties.properties) {
                redef.additionalProperties.properties = this._fixDefinitionPropertiesCollection(redef.additionalProperties.properties);
            }
            if (Array.isArray(redef.additionalProperties.type)) {
                redef.additionalProperties.type = redef.additionalProperties.type[0];
            }
        }
        else if (redef.additionalProperties === true && !this.opts.outputAdditionalPropertiesTrue) {
            delete redef.additionalProperties;
        }
        return redef;
    }
    _fixDefinitionPropertiesCollection(props) {
        if (!props) {
            return;
        }
        let removeProps = [];
        for (const property of Object.keys(props)) {
            const pdef = props[property];
            if (typeof pdef['$ref'] !== 'undefined') {
                pdef['$ref'] = this.replaceRefStr(pdef['$ref']);
                continue;
            }
            if (pdef.type === 'array') {
                if (pdef.items && pdef.items['$ref']) {
                    let items = [];
                    if (Array.isArray(pdef.items)) {
                        items = pdef.items;
                    }
                    else {
                        items.push(pdef.items);
                    }
                    pdef.items = this._fixDefinitionPropertiesCollection(items);
                    if (pdef.items.length === 0) {
                        pdef.items = pdef.items[0];
                    }
                }
            }
            else if (typeof pdef.type === 'string') {
                if (this.removableTypes.indexOf(pdef.type) !== -1) {
                    removeProps.push(property);
                }
            }
            else if (typeof pdef.type !== 'undefined') {
                let newTypes = pdef.type.filter(t => this.removableTypes.indexOf(t) === -1);
                if (newTypes.length === 0) {
                    removeProps.push(property);
                }
                else if (newTypes.length === 1) {
                    pdef.type = newTypes[0];
                }
                else {
                    pdef.oneOf = newTypes.map(t => {
                        if (typeof t === 'string') {
                            return {
                                type: t,
                            };
                        }
                        else {
                            return t;
                        }
                    });
                    delete pdef.type;
                    // (<OpenAPIV3.ArraySchemaObject>pdef).nullable = true;
                }
            }
            if (typeof pdef.additionalProperties === 'object' && Array.isArray(pdef.additionalProperties.type)) {
                if (this.opts.outputAdditionalPropertiesTrue) {
                    pdef.additionalProperties = true;
                }
                else {
                    pdef.additionalProperties.type = pdef.additionalProperties.type[0];
                }
            }
            props[property] = this.fixDefinitionProperties(pdef);
        }
        for (const prop of removeProps) {
            delete props[prop];
        }
        return props;
    }
    fixDefinitionRefs(def) {
        for (const property of Object.keys(def)) {
            if (property === '$ref') {
                def[property] = this.replaceRefStr(def[property]);
            }
            else {
                const val = def[property];
                if (typeof val === 'object') {
                    if (Array.isArray(val)) {
                        this.fixArrayDefinitionRefs(val);
                    }
                    else {
                        this.fixDefinitionRefs(val);
                    }
                }
            }
        }
        return def;
    }
    fixArrayDefinitionRefs(val) {
        val.forEach(v => {
            if (typeof v === 'object') {
                if (Array.isArray(v)) {
                    this.fixArrayDefinitionRefs(v);
                }
                else {
                    this.fixDefinitionRefs(v);
                }
            }
        });
    }
    replaceRefStr(refStr) {
        if (refStr.includes('/')) {
            refStr = refStr.substring(refStr.lastIndexOf('/') + 1);
        }
        const expected = this.renameDefinition(refStr);
        if (!this.definitions.has(expected)) {
            this.expectedDefinitions.add(refStr);
        }
        return `#/components/schemas/def___${expected}`;
    }
    recordTagObject(t) {
        if (!this.tags.has(t.name) || !this.tags.get(t.name).description) {
            this.tags.set(t.name, t);
        }
        return t.name;
    }
    getTagObject(t) {
        return {
            name: t.name,
            description: t.description,
        };
    }
    getParametersObject(p) {
        let inStr;
        let name = p.paramDef.propertyKey.toString();
        switch (p.paramDef.type) {
            case 10 /* ApiParamType.FormFileSingle */:
                const formFieldName = (0, TransformerMetadata_1.getMetadataValueByDescriptor)(p.metadata, TransformerMetadata_1.BuiltinMetadata.FormDataFieldName);
                if (formFieldName) {
                    name = formFieldName;
                }
            // Intentional flow-through
            case 0 /* ApiParamType.Body */:
                inStr = 'body';
                break;
            case 2 /* ApiParamType.Query */:
                inStr = 'query';
                break;
            case 3 /* ApiParamType.Path */:
                inStr = 'path';
                break;
            case 4 /* ApiParamType.Header */:
                inStr = 'header';
                break;
            case 5 /* ApiParamType.Callback */:
            case 6 /* ApiParamType.Transport */:
            case 7 /* ApiParamType.Dependency */:
            case 8 /* ApiParamType.Out */:
            case 9 /* ApiParamType.Custom */:
                return null;
            default:
                throw new Error(`Unknown Api Parameter Type: ${p.type}`);
        }
        let schema = this.getInlineTypeSchema(p.paramDef.args.typedef);
        let enumVals;
        if (p.paramDef.args.typedef.type === 'string' || p.paramDef.args.typedef.type === 'number' || p.paramDef.args.typedef.type === 'enum') {
            if (p.paramDef.args.typedef.schema) {
                if ('enum' in p.paramDef.args.typedef.schema) {
                    enumVals = p.paramDef.args.typedef.schema.enum;
                }
                else if ('const' in p.paramDef.args.typedef.schema) {
                    enumVals = [p.paramDef.args.typedef.schema.const];
                }
            }
        }
        return {
            name,
            in: inStr,
            required: this.isParameterRequired(p),
            description: p.paramDef.args.description,
            schema,
        };
    }
    isParameterRequired(p) {
        return !p.paramDef.args.optional && !p.paramDef.args.initializer;
    }
    getRequestBody(p) {
        if (p.find(p => p.paramDef.type === 10 /* ApiParamType.FormFileSingle */)) {
            if (p.find(p => p.paramDef.type !== 10 /* ApiParamType.FormFileSingle */)) {
                throw new Error('Multipart body types cannot be mixed with body params of other types.');
            }
            return this.getMultipartRequestBodySchema(p);
        }
        if (p.find(p => p.paramDef.type === 0 /* ApiParamType.Body */ || p.paramDef.type === 1 /* ApiParamType.RawBody */)) {
            // uses traditional body params
            const bodyParams = p.filter(p => p.paramDef.args.typedef || p.paramDef.args.typeref);
            if (bodyParams.length > 1) {
                throw new Error('Unsupported: multiple type-asserting body parameters');
            }
            if (bodyParams.length === 0) {
                if (p.length === 1) {
                    return this.getSingleRequestBody(p[0]);
                }
                return undefined;
            }
            return this.getSingleRequestBody(bodyParams[0]);
        }
        return undefined;
    }
    getMultipartRequestBodySchema(params) {
        const properties = {};
        const requiredProperties = [];
        for (const p of params) {
            let param;
            if (p.paramDef.type === 10 /* ApiParamType.FormFileSingle */) {
                const name = (0, TransformerMetadata_1.getMetadataValueByDescriptor)(p.metadata, TransformerMetadata_1.BuiltinMetadata.FormDataFieldName);
                if (properties[name]) {
                    throw new Error('Multiple form files are not yet supported');
                }
                param = {
                    name,
                    in: 'body',
                    required: this.isParameterRequired(p),
                    description: p.paramDef.args.description,
                    schema: {
                        type: 'string',
                        format: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(p.metadata, TransformerMetadata_1.BuiltinMetadata.SchemaFormat) ?? 'binary',
                    },
                };
            }
            else {
                param = this.getParametersObject(p);
            }
            if (param.required) {
                requiredProperties.push(param.name);
            }
            properties[param.name] = {
                description: param.description,
                ...(param.schema),
            };
        }
        return {
            content: {
                [apiManagement_1.ApiMimeType.MultipartFormData]: {
                    schema: {
                        type: 'object',
                        properties,
                        required: requiredProperties.length > 0 ? requiredProperties : undefined,
                    }
                }
            }
        };
        ;
    }
    getSingleRequestBody(p) {
        let schema;
        if (p.paramDef.args.typedef) {
            schema = this.getInlineTypeSchema(p.paramDef.args.typedef);
            let enumVals;
            if (p.paramDef.args.typedef.type === 'string' || p.paramDef.args.typedef.type === 'number' || p.paramDef.args.typedef.type === 'enum') {
                if (p.paramDef.args.typedef.schema) {
                    if ('enum' in p.paramDef.args.typedef.schema) {
                        enumVals = p.paramDef.args.typedef.schema.enum;
                    }
                    else if ('const' in p.paramDef.args.typedef.schema) {
                        enumVals = [p.paramDef.args.typedef.schema.const];
                    }
                }
            }
        }
        else if (p.paramDef.type === 1 /* ApiParamType.RawBody */) {
            schema = {
                type: 'string',
                format: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(p.metadata, TransformerMetadata_1.BuiltinMetadata.SchemaFormat) ?? 'binary',
            };
        }
        else {
            return undefined;
        }
        const mimeType = ((0, TransformerMetadata_1.getMetadataValueByDescriptor)(p.metadata, TransformerMetadata_1.BuiltinMetadata.MimeType)
            ?? apiManagement_1.ApiMimeType.ApplicationJson);
        return {
            required: this.isParameterRequired(p),
            description: p.paramDef.args.description,
            content: {
                [mimeType]: {
                    schema,
                }
            }
        };
    }
    toString() {
        return yaml.safeDump(this.getDocument(), {
            ...this.opts.yamlOpts,
            skipInvalid: true,
        });
    }
}
exports.OpenApiV3Extractor = OpenApiV3Extractor;
OpenApiV3Extractor.SwaggerVersion = '3.0.1';
OpenApiV3Extractor.RouteSeperator = '/';
OpenApiV3Extractor.MimeTypeText = 'text/plain';
OpenApiV3Extractor.MimeTypeJson = 'application/json';
OpenApiV3Extractor.ExcludeMetadataTags = ['private'];
class OpenApiV3JsonExtractor extends OpenApiV3Extractor {
    toString() {
        return JSON.stringify(this.getDocument(), undefined, 4);
    }
}
exports.OpenApiV3JsonExtractor = OpenApiV3JsonExtractor;
//# sourceMappingURL=OpenApiV3.js.map