"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Swagger2JsonExtractor = exports.Swagger2Extractor = void 0;
const TransformerMetadata_1 = require("../transformer/TransformerMetadata");
const yaml = require("js-yaml");
const HandlerTree_1 = require("../transformer/HandlerTree");
const apiManagement_1 = require("../apiManagement");
class Swagger2Extractor {
    constructor(apiTree, apiInfo, opts) {
        this.apiTree = apiTree;
        this.apiInfo = apiInfo;
        this.opts = opts;
        this.tags = new Map();
        this.definitions = new Map();
    }
    getDocument() {
        const doc = {
            swagger: Swagger2Extractor.SwaggerVersion,
            info: {
                title: this.apiInfo.title,
                version: this.apiInfo.version,
                description: this.apiInfo.description,
                license: this.apiInfo.license ? this.apiInfo.license[0] : undefined,
                termsOfService: this.apiInfo.termsOfService,
                contact: this.apiInfo.contact,
            },
            host: this.validHost(this.apiInfo.host || this.apiInfo.homepage),
            basePath: this.apiInfo.basePath,
            schemes: this.apiInfo.schemes || this.tryInferSchemes(),
            paths: this.getPaths(),
        };
        if (this.tags.size > 0) {
            doc.tags = [];
            this.tags.forEach(tag => doc.tags.push(this.getTagObject(tag)));
        }
        if (this.definitions.size > 0) {
            doc.definitions = this.getDefinitions();
        }
        return doc;
    }
    getDefinitions() {
        const defKeys = Array.from(this.definitions.keys());
        defKeys.sort();
        const definitions = {};
        defKeys.forEach(defName => {
            definitions[defName] = this.definitions.get(defName);
        });
        return definitions;
    }
    tryInferSchemes() {
        if (this.opts.disableTryInferSchemes) {
            return;
        }
        const schemes = new Set();
        if (this.apiInfo.host) {
            const url = new URL(this.apiInfo.host);
            if (url.protocol.length > 0 && !schemes.has(url.protocol)) {
                schemes.add(url.protocol.replace(':', ''));
            }
        }
        if (schemes.size > 0) {
            return Array.from(schemes);
        }
    }
    validHost(host) {
        try {
            const url = new URL(host);
            if (url.host.length > 0) {
                return url.host;
            }
        }
        catch (e) {
        }
        return host;
    }
    getPaths() {
        const paths = {};
        // Array.from(WalkChildrenByType(api, isHandlerParameterNode)).forEach(api => {
        for (const api of (0, HandlerTree_1.WalkTreeByType)(this.apiTree, HandlerTree_1.isHandlerNode)) {
            const operation = this.getOperationObject(api);
            if (!operation) {
                continue;
            }
            const route = this.swaggerizeRoute(api.route);
            if (!paths[route]) {
                paths[route] = {};
            }
            if (!paths[route][api.apiMethod.toLowerCase()]) {
                paths[route][api.apiMethod.toLowerCase()] = operation;
            }
            else {
                throw new Error(`Multiple APIs for route: [${api.apiMethod}]: ${route}`);
            }
        }
        ;
        return paths;
    }
    swaggerizeRoute(route) {
        return apiManagement_1.ManagedApi.GetRouteTokens(route)
            .map(t => {
            if (typeof t === 'string') {
                return t;
            }
            let outStr = t.prefix ? t.prefix : '';
            if (t.modifier && t.modifier !== '?') {
                console.warn('Unhandled modifier: ' + t.modifier);
            }
            outStr += `{${t.name}${t.modifier ? t.modifier : ''}}`;
            if (t.suffix) {
                outStr += t.suffix;
            }
            return outStr;
        })
            .join('');
    }
    getOperationObject(api) {
        if ((0, TransformerMetadata_1.getMetadataValue)(api.metadata, "OpenApi" /* IMetadataType.OpenApi */, undefined, "private" /* OpenApiMetadataType.Private */)) {
            return;
        }
        const metadataTags = (0, TransformerMetadata_1.getAllMetadataValues)(api.metadata, "OpenApi" /* IMetadataType.OpenApi */, undefined, "tag" /* OpenApiMetadataType.Tag */);
        return {
            operationId: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(api.metadata, TransformerMetadata_1.BuiltinMetadata.Name),
            description: (0, TransformerMetadata_1.getMetadataValue)(api.metadata, "OpenApi" /* IMetadataType.OpenApi */, undefined, "description" /* OpenApiMetadataType.Description */),
            summary: (0, TransformerMetadata_1.getMetadataValue)(api.metadata, "OpenApi" /* IMetadataType.OpenApi */, undefined, "summary" /* OpenApiMetadataType.Summary */),
            tags: metadataTags.map(t => this.recordTagObject(t)),
            parameters: Array.from((0, HandlerTree_1.WalkChildrenByType)(api, HandlerTree_1.isHandlerParameterNode)).map(p => this.getParametersObject(p)),
            responses: {
                default: {
                    schema: this.getInlineTypeSchema((0, TransformerMetadata_1.getMetadataValueByDescriptor)(api.metadata, TransformerMetadata_1.BuiltinMetadata.ReturnSchema)),
                    description: this.getOperationResponseDescription(api),
                }
            }
        };
    }
    getOperationResponseDescription(api) {
        const metadata = (0, TransformerMetadata_1.getMetadataValue)(api.metadata, "OpenApi" /* IMetadataType.OpenApi */, undefined, "response" /* OpenApiMetadataType.ResponseDescription */);
        if (metadata) {
            return metadata;
        }
        const schema = (0, TransformerMetadata_1.getMetadataValueByDescriptor)(api.metadata, TransformerMetadata_1.BuiltinMetadata.ReturnSchema);
        if (schema) {
            if (schema.typename) {
                return schema.typename;
            }
            return `${schema.type} response`;
        }
        return 'Default response';
    }
    getInlineTypeSchema(returnType) {
        switch (returnType.type) {
            case 'object':
                if (returnType.schema) {
                    return this.getInternalSchemaToOutputSchema(returnType.schema);
                }
            // Fall through to next case is intentional
            case 'number':
            case 'string':
            case 'boolean':
                return {
                    type: returnType.type,
                };
            case 'void':
                return {};
            default:
                throw new Error(`Unable to serialize inline type: ${returnType.typename} (type: ${returnType.type})`);
        }
    }
    getInternalSchemaToOutputSchema(schema) {
        if (schema.definitions) {
            this.addDefinitions(schema.definitions);
            schema = { ...schema };
            delete schema.definitions;
        }
        if (schema.$schema) {
            delete schema.$schema;
        }
        return schema;
    }
    addDefinitions(definitions) {
        for (const definition of Object.keys(definitions)) {
            if (!this.definitions.has(definition)) {
                this.definitions.set(definition, definitions[definition]);
            }
        }
    }
    recordTagObject(t) {
        if (!this.tags.has(t.name) || !this.tags.get(t.name).description) {
            this.tags.set(t.name, t);
        }
        return t.name;
    }
    getTagObject(t) {
        return {
            name: t.name,
            description: t.description,
        };
    }
    getParametersObject(p) {
        let inStr;
        switch (p.paramDef.type) {
            case 0 /* ApiParamType.Body */:
                inStr = 'body';
                break;
            case 2 /* ApiParamType.Query */:
                inStr = 'query';
                break;
            case 3 /* ApiParamType.Path */:
                inStr = 'path';
                break;
            default:
                return;
        }
        let schema;
        if (p.paramDef.args.typedef.type === 'object') {
            schema = this.getInlineTypeSchema(p.paramDef.args.typedef);
        }
        let enumVals;
        if (p.paramDef.args.typedef.type === 'string' || p.paramDef.args.typedef.type === 'number' || p.paramDef.args.typedef.type === 'enum') {
            if (p.paramDef.args.typedef.schema) {
                if ('enum' in p.paramDef.args.typedef.schema) {
                    enumVals = p.paramDef.args.typedef.schema.enum;
                }
                else if ('const' in p.paramDef.args.typedef.schema) {
                    enumVals = [p.paramDef.args.typedef.schema.const];
                }
            }
        }
        return {
            name: p.paramDef.propertyKey.toString(),
            in: inStr,
            required: !p.paramDef.args.optional && !p.paramDef.args.initializer,
            description: p.paramDef.args.description,
            type: p.paramDef.args.typedef.type,
            schema,
            enum: enumVals,
        };
    }
    toString() {
        return yaml.safeDump(this.getDocument(), {
            ...this.opts.yamlOpts,
            skipInvalid: true,
        });
    }
}
exports.Swagger2Extractor = Swagger2Extractor;
Swagger2Extractor.SwaggerVersion = '2.0';
Swagger2Extractor.RouteSeperator = '/';
Swagger2Extractor.ExcludeMetadataTags = ['private'];
class Swagger2JsonExtractor extends Swagger2Extractor {
    toString() {
        return JSON.stringify(this.getDocument(), undefined, 4);
    }
}
exports.Swagger2JsonExtractor = Swagger2JsonExtractor;
//# sourceMappingURL=Swagger2Extractor.js.map