"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OutputFileGenerator = void 0;
const path = require("path");
const fs = require("promise-fs");
const parseIgnore = require("parse-gitignore");
const AsyncGlob_1 = require("../Util/AsyncGlob");
class OutputFileGenerator {
    constructor(baseDir) {
        this.outFiles = new Map();
        if (path.isAbsolute(baseDir)) {
            this.baseDir = baseDir;
        }
        else {
            this.baseDir = path.join(process.cwd(), baseDir);
        }
    }
    addOutputFile(filePath, generatorFunc) {
        if (this.outFiles.has(filePath)) {
            throw new Error(`Output file already registered: ${filePath}`);
        }
        if (path.isAbsolute(filePath)) {
            throw new Error('Output file paths must be relative to the root directory');
        }
        this.outFiles.set(filePath, {
            path: filePath,
            generatorFunc,
        });
    }
    async generate() {
        if (fs.existsSync(this.baseDir)) {
            const stat = await fs.lstat(this.baseDir);
            if (!stat.isDirectory) {
                throw new Error('Cannot set a file as the output directory');
            }
            if (!this.untouchableFiles) {
                await this.loadUntouchableFiles();
            }
            await this.clearOutputDirectory();
        }
        else {
            await fs.mkdir(this.baseDir);
        }
        const writtenFiles = [];
        for (const file of this.outFiles.values()) {
            const outPath = path.join(this.baseDir, file.path);
            if (await this.generateFile(outPath, file)) {
                writtenFiles.push(outPath);
            }
        }
        return writtenFiles;
    }
    async generateFile(outPath, file) {
        if (this.isUntouchable(outPath)) {
            return false;
        }
        const dirName = path.dirname(outPath);
        const outBuffer = await file.generatorFunc(outPath);
        if (!outBuffer) {
            return false;
        }
        if (!fs.existsSync(dirName)) {
            await this.createPath(dirName);
        }
        await fs.writeFile(outPath, outBuffer);
        return true;
    }
    async createPath(dirname) {
        const parentName = path.dirname(dirname);
        if (!fs.existsSync(parentName)) {
            await this.createPath(parentName);
        }
        await fs.mkdir(dirname);
    }
    isUntouchable(file) {
        if (!this.untouchableFiles) {
            return false;
        }
        if (!path.isAbsolute(file)) {
            file = path.join(this.baseDir, file);
        }
        return this.untouchableFiles.has(file);
    }
    async loadUntouchableFiles() {
        const ignoreFile = this.getIgnoreFilePath();
        if (!fs.existsSync(ignoreFile)) {
            this.untouchableFiles = new Set();
            return;
        }
        const patterns = parseIgnore(await fs.readFile(ignoreFile, { encoding: this.getDefaultEncoding() }));
        this.untouchableFiles = new Set((await Promise.all(patterns
            .map(p => this.matchPatternInOutputDir(p))))
            .reduce((prev, curr) => prev.concat(curr), [])
            .map(f => path.resolve(this.baseDir, f)));
        this.untouchableFiles.add(ignoreFile);
    }
    matchPatternInOutputDir(p) {
        return (0, AsyncGlob_1.asyncGlob)(p, {
            cwd: this.baseDir,
        });
    }
    getIgnoreFilePath() {
        return path.join(this.baseDir, OutputFileGenerator.IGNORE_FILE);
    }
    getDefaultEncoding() {
        return 'utf8';
    }
    async clearOutputDirectory(dir = this.baseDir) {
        let hadUntouchableFiles = false;
        for (const fileName of await fs.readdir(dir)) {
            const file = path.join(dir, fileName);
            if (this.isUntouchable(file)) {
                hadUntouchableFiles = true;
                continue;
            }
            const fileStat = await fs.lstat(file);
            if (fileStat.isDirectory()) {
                if (!await this.clearOutputDirectory(file)) {
                    await fs.rmdir(file);
                }
            }
            else if (fileStat.isFile()) {
                await fs.unlink(file);
            }
        }
        return hadUntouchableFiles;
    }
}
exports.OutputFileGenerator = OutputFileGenerator;
OutputFileGenerator.IGNORE_FILE = '.tsapi-generator-ignore';
//# sourceMappingURL=OutputFileGenerator.js.map