"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BuiltinArgumentExtractors = void 0;
const ts = require("typescript");
const InternalTypes_1 = require("../apiManagement/InternalTypes");
const TransformerMetadata_1 = require("./TransformerMetadata");
const ExpressionWrapper_1 = require("./ExpressionWrapper");
const ApiProcessing_1 = require("../apiManagement/ApiProcessing/ApiProcessing");
const Constants_1 = require("../Constants");
class BuiltinArgumentExtractors {
    static GetNodeReturnType(node, context) {
        if (ts.isMethodDeclaration(node)) {
            const type = context.typeChecker.getTypeAtLocation(node);
            const callSignatures = type.getCallSignatures();
            if (callSignatures.length === 1) {
                return context.typeSerializer.getInternalTypeRepresentation(node.type, context.typeChecker.getReturnTypeOfSignature(callSignatures[0]));
            }
            else if (callSignatures.length > 1) {
                throw new Error('Cannot handle method with multiple call signatures');
            }
        }
    }
    static GetTypeArgumentType(node, context, argNo) {
        if (ts.isDecorator(node) && ts.isCallExpression(node.expression)) {
            if (node.expression.typeArguments.length <= argNo) {
                return undefined;
            }
            const type = context.typeChecker.getTypeAtLocation(node.expression.typeArguments[argNo]);
            return context.typeSerializer.getInternalTypeRepresentation(node.expression.typeArguments[argNo], type);
        }
        return undefined;
    }
    static parenthesizeExpression(expression) {
        return ts.isParenthesizedExpression(expression)
            ? expression
            : ts.factory.createParenthesizedExpression(expression);
    }
}
exports.BuiltinArgumentExtractors = BuiltinArgumentExtractors;
BuiltinArgumentExtractors.RouteArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeString,
    metadataExtractor: (args) => ({
        ...TransformerMetadata_1.BuiltinMetadata.Route,
        value: args.transformContext.typeSerializer.compileExpressionToStringConstant(args.argumentExpression),
    }),
};
BuiltinArgumentExtractors.ResponseCodesArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeNumberArray,
    defaultExpression: ts.factory.createArrayLiteralExpression([ts.factory.createNumericLiteral(Constants_1.DefaultApiResponseCode.toString())]),
    metadataExtractor: (args) => ({
        ...TransformerMetadata_1.BuiltinMetadata.ResponseCodes,
        value: args.transformContext.typeSerializer.compileExpressionToArrayConstant(args.argumentExpression),
    }),
};
BuiltinArgumentExtractors.ApiProcessorStageArgument = {
    type: {
        ...InternalTypes_1.InternalTypeUtil.TypeEnum,
        schema: { enum: [ApiProcessing_1.ApiProcessorTime.StagePreInvoke, ApiProcessing_1.ApiProcessorTime.StagePostInvoke] },
    },
    metadataExtractor: (args) => ({
        ...TransformerMetadata_1.BuiltinMetadata.ApiProcessorStage,
        value: args.transformContext.typeSerializer.compileExpressionToStringConstant(args.argumentExpression),
    }),
};
BuiltinArgumentExtractors.ApiProcessorScopeArgument = {
    optional: true,
    type: {
        ...InternalTypes_1.InternalTypeUtil.TypeEnum,
        schema: { enum: [ApiProcessing_1.ApiProcessorScope.ScopeGlobal, ApiProcessing_1.ApiProcessorScope.ScopeClass] },
    },
    metadataExtractor: (args) => ({
        ...TransformerMetadata_1.BuiltinMetadata.ApiProcessorScope,
        value: args.transformContext.typeSerializer.compileExpressionToStringConstant(args.argumentExpression),
    }),
};
BuiltinArgumentExtractors.NameArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeString,
    metadataExtractor: (args) => ({
        ...TransformerMetadata_1.BuiltinMetadata.Name,
        value: args.transformContext.typeSerializer.compileExpressionToStringConstant(args.argumentExpression),
    }),
};
BuiltinArgumentExtractors.FunctionArgumentNameArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeString,
    optional: false,
    metadataExtractor: (args) => ({
        ...TransformerMetadata_1.BuiltinMetadata.FunctionArgumentName,
        value: args.transformContext.typeSerializer.compileExpressionToStringConstant(args.argumentExpression),
    }),
};
BuiltinArgumentExtractors.SchemaOutputRefOverrideArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeString,
    optional: true,
    metadataExtractor: (args) => ({
        ...TransformerMetadata_1.BuiltinMetadata.SchemaRefOverride,
        value: args.transformContext.typeSerializer.compileExpressionToStringConstant(args.argumentExpression),
    }),
};
BuiltinArgumentExtractors.OptionalNameArgument = {
    ...BuiltinArgumentExtractors.NameArgument,
    optional: true,
};
BuiltinArgumentExtractors.MimeTypeArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeString,
    metadataExtractor: (args) => ({
        ...TransformerMetadata_1.BuiltinMetadata.MimeType,
        value: args.transformContext.typeSerializer.compileExpressionToStringConstant(args.argumentExpression),
    }),
};
BuiltinArgumentExtractors.OptionalMimeTypeArgument = {
    ...BuiltinArgumentExtractors.MimeTypeArgument,
    optional: true,
};
BuiltinArgumentExtractors.DependencyScopeArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeString,
    optional: true,
    metadataExtractor: (args) => ({
        ...TransformerMetadata_1.BuiltinMetadata.DependencyScope,
        value: args.transformContext.typeSerializer.compileExpressionToStringConstant(args.argumentExpression),
    }),
};
BuiltinArgumentExtractors.RegexpArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeRegex,
    optional: true,
    metadataExtractor: (args) => {
        if (!ts.isRegularExpressionLiteral(args.argumentExpression)) {
            throw new Error('Regular expression must be a regex literal');
        }
        return {
            ...TransformerMetadata_1.BuiltinMetadata.ValidationRegExp,
            value: new ExpressionWrapper_1.ExpressionWrapper(args.argumentExpression),
        };
    },
};
BuiltinArgumentExtractors.NumberMinArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeNumber,
    optional: true,
    metadataExtractor: (args) => {
        return {
            ...TransformerMetadata_1.BuiltinMetadata.NumberMin,
            value: args.transformContext.typeSerializer.compileExpressionToNumericConstant(args.argumentExpression),
        };
    },
};
BuiltinArgumentExtractors.NumberMaxArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeNumber,
    optional: true,
    metadataExtractor: (args) => {
        return {
            ...TransformerMetadata_1.BuiltinMetadata.NumberMax,
            value: args.transformContext.typeSerializer.compileExpressionToNumericConstant(args.argumentExpression),
        };
    },
};
BuiltinArgumentExtractors.BoolUndefinedIfOmitted = {
    type: InternalTypes_1.InternalTypeUtil.TypeBoolean,
    optional: true,
    metadataExtractor: (args) => {
        return {
            ...TransformerMetadata_1.BuiltinMetadata.UndefinedIfOmitted,
            value: args.transformContext.typeSerializer.compileExpressionToNumericConstant(args.argumentExpression),
        };
    },
};
BuiltinArgumentExtractors.ValidationFunctionArgument = {
    type: InternalTypes_1.InternalTypeUtil.TypeAnyFunction,
    optional: true,
    metadataExtractor: (args) => {
        return {
            ...TransformerMetadata_1.BuiltinMetadata.ValidationFunction,
            value: new ExpressionWrapper_1.ExpressionWrapper(BuiltinArgumentExtractors.parenthesizeExpression(args.argumentExpression)),
        };
    },
};
BuiltinArgumentExtractors.ReturnSchemaArgument = {
    type: { type: 'any' },
    optional: true,
    metadataExtractor: (args) => {
        let returnType = BuiltinArgumentExtractors.GetNodeReturnType(args.node, args.transformContext);
        if (returnType) {
            return {
                ...TransformerMetadata_1.BuiltinMetadata.ReturnSchema,
                value: returnType
            };
        }
    },
    transformer: (args) => {
        let returnType = BuiltinArgumentExtractors.GetNodeReturnType(args.node, args.transformContext);
        if (returnType) {
            return args.transformContext.typeSerializer.objectToLiteral(returnType);
        }
    }
};
BuiltinArgumentExtractors.DecoratorTypeArgTypeArgument = (argNo) => ({
    type: { type: 'any' },
    optional: true,
    metadataExtractor: (args) => {
        if (!ts.canHaveDecorators(args.node)) {
            return;
        }
        const decorator = args.node.modifiers?.find(d => d.kind === ts.SyntaxKind.Decorator && d.expression?.expression?.escapedText === args.decorator.magicFunctionName);
        let returnType = BuiltinArgumentExtractors.GetTypeArgumentType(decorator, args.transformContext, argNo);
        if (returnType) {
            return {
                ...TransformerMetadata_1.BuiltinMetadata.DecoratorTypeArgType,
                value: returnType
            };
        }
    },
    transformer: (args) => {
        if (!ts.canHaveDecorators(args.node)) {
            return;
        }
        const decorator = args.node.modifiers?.find(d => d.kind === ts.SyntaxKind.Decorator && d.expression?.expression?.escapedText === args.decorator.magicFunctionName);
        let returnType = BuiltinArgumentExtractors.GetTypeArgumentType(decorator, args.transformContext, argNo);
        if (returnType) {
            return args.transformContext.typeSerializer.objectToLiteral(returnType);
        }
    }
});
//# sourceMappingURL=BuiltinArgumentExtractors.js.map