"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Decorator = exports.DecoratorNodeType = void 0;
const ts = require("typescript");
const path = require("path");
const DecoratorDefinition_1 = require("./DecoratorDefinition");
const TransformerMetadata_1 = require("./TransformerMetadata");
const TransformerUtil_1 = require("./TransformerUtil");
const ExpressionWrapper_1 = require("./ExpressionWrapper");
const CompilationError_1 = require("../Util/CompilationError");
var DecoratorNodeType;
(function (DecoratorNodeType) {
    DecoratorNodeType[DecoratorNodeType["Class"] = 0] = "Class";
    DecoratorNodeType[DecoratorNodeType["ClassProperty"] = 1] = "ClassProperty";
    DecoratorNodeType[DecoratorNodeType["Method"] = 2] = "Method";
    DecoratorNodeType[DecoratorNodeType["Parameter"] = 3] = "Parameter";
})(DecoratorNodeType || (exports.DecoratorNodeType = DecoratorNodeType = {}));
class Decorator {
    get magicFunctionName() {
        return this.definition.magicFunctionName;
    }
    get indexTs() {
        return this._indexTs;
    }
    get provider() {
        return this.definition.provider;
    }
    get decoratorType() {
        return this.definition.decoratorType;
    }
    get dependencies() {
        return this.definition.dependencies;
    }
    get arguments() {
        return this.definition.arguments;
    }
    get metadata() {
        return this.definition.metadata || [];
    }
    get transformArgumentsToObject() {
        if (typeof this.definition.transformArgumentsToObject === 'undefined') {
            return false;
        }
        return this.definition.transformArgumentsToObject;
    }
    get isCallExpression() {
        if (typeof this.definition.isCallExpression === 'boolean') {
            return this.definition.isCallExpression;
        }
        return true;
    }
    get treeHierarchyType() {
        if (typeof this.definition.treeHierarchyType !== 'undefined') {
            return this.definition.treeHierarchyType;
        }
        return DecoratorDefinition_1.DecoratorNodeTreeHierarchyType.Child;
    }
    constructor(definition, nodeType) {
        this.definition = definition;
        this.nodeType = nodeType;
        this._indexTs = this.getPathWithoutCommonRoot(this.getPathWithoutExt(this.definition.indexTs), __dirname);
    }
    getPathWithoutExt(p) {
        const ext = path.extname(p);
        return path.join(path.dirname(p), path.basename(p, ext));
    }
    getPathWithoutCommonRoot(path, common) {
        const root = this.getPathFromCommonRoot(path, common);
        return path.substr(root.length);
    }
    getPathFromCommonRoot(a, b) {
        let i;
        for (i = 0; i < a.length && i < b.length; ++i) {
            if (a[i] !== b[i]) {
                break;
            }
        }
        return a.substr(0, i);
    }
    getExt(p) {
        if (p.endsWith('.d.ts')) {
            return '.d.ts';
        }
        else {
            return path.extname(p);
        }
    }
    getNodeLocation(node) {
        const sourceFile = node.getSourceFile();
        const posDetails = sourceFile.getLineAndCharacterOfPosition(node.pos);
        return {
            ...posDetails,
            file: sourceFile.fileName,
            position: node.pos,
        };
    }
    getMetadata(node, decorator, context) {
        const metadata = [...this.metadata];
        for (const getter of this.getDefaultMetadataGetters()) {
            const meta = getter(node, decorator, context);
            if (meta) {
                metadata.push(meta);
            }
        }
        return metadata;
    }
    *getDefaultMetadataGetters() {
        yield (node, decorator, context) => this.getNodeNameMetadata(node);
        yield (node, decorator, context) => this.getDecoratorTypeArgumentsMetadata(node, decorator);
    }
    getNodeNameMetadata(node) {
        if ((0, TransformerUtil_1.isNamedNode)(node)) {
            let name;
            if (typeof node.name === 'string') {
                name = node.name;
            }
            else if (ts.isIdentifier(node.name)) {
                name = node.name.text;
            }
            else {
                throw new CompilationError_1.CompilationError('Unknown node name type', node);
            }
            return {
                ...TransformerMetadata_1.BuiltinMetadata.Name,
                value: name,
            };
        }
    }
    getDecoratorTypeArgumentsMetadata(node, decorator) {
        // return BuiltinArgumentExtractors.DecoratorTypeArgTypeArgument.metadataExtractor()
        return undefined;
    }
    applyArguments(node, decorator, context) {
        let metadata = this.getMetadata(node, decorator, context);
        if (!ts.isCallExpression(decorator.expression)) {
            return {
                metadata,
                decorator,
            };
        }
        const exprArguments = [];
        for (let index = 0; index < this.arguments.length; ++index) {
            const argDef = this.arguments[index];
            let arg = this.getUnparenthesizedExpression(decorator.expression.arguments[index]);
            const isUndefined = typeof arg === 'undefined'
                || arg.kind === ts.SyntaxKind.NullKeyword
                || arg.kind === ts.SyntaxKind.UndefinedKeyword
                || (ts.isIdentifier(arg) && ts.identifierToKeywordKind(arg) === ts.SyntaxKind.UndefinedKeyword);
            if (isUndefined) {
                if (argDef.defaultExpression) {
                    arg = argDef.defaultExpression;
                }
                else if (argDef.optional) {
                    if (!argDef.transformer) {
                        continue;
                    }
                }
                else if (!argDef.optional) {
                    throw new Error('Expected argument');
                }
            }
            else {
                this.checkTypeCompatibility(argDef, arg, context);
            }
            exprArguments[index] = { expression: arg };
            if (argDef.metadataExtractor || argDef.transformer) {
                const processorArgs = {
                    node,
                    argument: argDef,
                    argumentExpression: arg,
                    index,
                    transformContext: context,
                    decorator: this,
                };
                let meta;
                if (argDef.metadataExtractor) {
                    meta = argDef.metadataExtractor(processorArgs) || undefined;
                    if (meta) {
                        metadata.push(meta);
                    }
                }
                if (argDef.transformer) {
                    let transformed = argDef.transformer(processorArgs);
                    if (transformed) {
                        exprArguments[index] = {
                            expression: transformed,
                            metadata: meta,
                        };
                    }
                }
                else if (meta) {
                    exprArguments[index] = {
                        expression: arg,
                        metadata: meta,
                    };
                }
            }
        }
        return {
            metadata,
            decorator: ts.factory.createDecorator(this.getOutputDecoratorArgs(decorator.expression, exprArguments, metadata, context)),
        };
    }
    getUnparenthesizedExpression(arg0) {
        if (!arg0) {
            return arg0;
        }
        while (ts.isParenthesizedExpression(arg0)) {
            arg0 = arg0.expression;
        }
        return arg0;
    }
    checkTypeCompatibility(argDef, arg, context) {
        // if (!ts.isTypeNode(arg)) {
        // 	throw new Error();
        // }
        // const type = context.typeChecker.getTypeFromTypeNode(arg.type);
    }
    getOutputDecoratorArgs(expression, exprArguments, metadataCollection, context) {
        if (this.definition.transformArgumentsToObject) {
            let paramArgs = {};
            if (Array.isArray(this.definition.transformArgumentsToObject)) {
                for (const descr of this.definition.transformArgumentsToObject) {
                    if (!descr.key) {
                        throw new Error('When transforming param args to object, descriptor must have key set');
                    }
                    let exprResult = (0, TransformerMetadata_1.getMetadataValueByDescriptor)(exprArguments.map(a => a.metadata), descr);
                    if (exprResult) {
                        if (exprResult.expression) {
                            paramArgs[descr.key] = new ExpressionWrapper_1.ExpressionWrapper(exprResult.expression);
                        }
                        else {
                            paramArgs[descr.key] = exprResult.metadata;
                        }
                    }
                    else {
                        paramArgs[descr.key] = (0, TransformerMetadata_1.getMetadataValueByDescriptor)(metadataCollection, descr);
                    }
                }
            }
            else {
                for (const metadata of metadataCollection) {
                    if (!metadata.key) {
                        continue;
                    }
                    paramArgs[metadata.key] = metadata.value;
                }
            }
            expression = ts.factory.createCallExpression(expression.expression, expression.typeArguments, ts.factory.createNodeArray([context.typeSerializer.objectToLiteral(paramArgs)]));
        }
        else {
            expression = ts.factory.createCallExpression(expression.expression, expression.typeArguments, ts.factory.createNodeArray(exprArguments.filter(arg => !!arg.expression).map(arg => arg.expression)));
        }
        return expression;
    }
    parenthesizeExpression(expression) {
        return ts.isParenthesizedExpression(expression)
            ? expression
            : ts.factory.createParenthesizedExpression(expression);
    }
    isSourceFileMatch(sourceFile) {
        let sourceFilePaths = [path.join(sourceFile.fileName)];
        if (sourceFilePaths[0] == this.indexTs) {
            return true;
        }
        const sourceFileExt = this.getExt(sourceFilePaths[0]);
        let parts = ['/dist/', '/src/'];
        if (process.env["DEBUG_PATH_REPLACE"]) {
            parts = process.env["DEBUG_PATH_REPLACE"].split(":");
            if (parts.length !== 2) {
                throw new Error("DEBUG_PATH_REPLACE must split into two parts using ':' separator");
            }
        }
        sourceFilePaths.push(sourceFilePaths[0].replace(parts[0], parts[1]));
        switch (sourceFileExt) {
            case '.d.ts':
                return !!sourceFilePaths.find(p => p.endsWith(this.indexTs + '.d.ts'));
            case '.ts':
                return !!sourceFilePaths.find(p => p.endsWith(this.indexTs + '.ts'));
            case '.js':
                return !!sourceFilePaths.find(p => p.endsWith(this.indexTs + '.js'));
            default:
                return false;
        }
    }
}
exports.Decorator = Decorator;
//# sourceMappingURL=Decorator.js.map