"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const DecoratorTransformer_1 = require("./DecoratorTransformer");
const ParamDecoratorTransformer_1 = require("./ParamDecoratorTransformer");
class ExtractionTransformer extends DecoratorTransformer_1.DecoratorTransformer {
    constructor(program, generator, transformInfo, onApiMethodExtracted, metadataManager, extractOnly = false) {
        super(program, generator, Object.assign({}, transformInfo, { nodeCheckFunction: (node => this.isDecoratedMethodDeclaration(node)) }));
        this.onApiMethodExtracted = onApiMethodExtracted;
        this.metadataManager = metadataManager;
        this.extractOnly = extractOnly;
        this.paramTransformers = new Map();
        this.transformInfo.parameterTypes.forEach(p => {
            this.paramTransformers.set(p, new ParamDecoratorTransformer_1.ParamDecoratorTransformer(this.program, this.generator, p));
        });
    }
    visitNode(node) {
        // Parse argument name
        let name;
        if (typeof node.name === 'string') {
            name = node.name;
        }
        else if (ts.isIdentifier(node.name)) {
            name = node.name.text;
        }
        else {
            throw new Error('Unknown node name type');
        }
        // Pull out the registered decorators
        const registeredDecorators = [this.transformInfo, ...this.metadataManager.getApiMethodMetadataDecorators()];
        const decoratorMap = new Map();
        for (const decorator of node.decorators) {
            for (const transformInfo of registeredDecorators) {
                if (this.isArgumentDecoratorCallExpression(decorator.expression)) {
                    decoratorMap.set(transformInfo, decorator.expression);
                }
            }
        }
        let definition;
        if (decoratorMap.has(this.transformInfo)) {
            const expression = decoratorMap.get(this.transformInfo);
            const type = this.typeChecker.getTypeAtLocation(node);
            const callSignatures = type.getCallSignatures();
            let returnType;
            if (callSignatures.length === 1) {
                returnType = this.typeSerializer.getInternalTypeRepresentation(node.type, this.typeChecker.getReturnTypeOfSignature(callSignatures[0]));
            }
            else if (callSignatures.length > 1) {
                throw new Error('Cannot handle method with multiple call signatures');
            }
            const nodeParams = this.parseApiMethodCallParameters(node.parameters);
            const result = this.getDecoratorArguments({
                route: null,
                handlerKey: name,
                method: this.transformInfo.apiDecoratorMethod,
                file: node.getSourceFile().fileName,
                parameters: nodeParams.paramDefs,
                returnType,
            }, expression);
            definition = result.definition;
            if (!this.extractOnly) {
                if (result.exprArguments) {
                    expression.arguments = result.exprArguments;
                }
                if (nodeParams.parameters) {
                    node.parameters = nodeParams.parameters;
                }
            }
        }
        if (definition) {
            let metadata = [];
            for (const decorator of decoratorMap) {
                metadata = metadata.concat(this.metadataManager.getApiMetadataForApiMethod(definition, node, decorator[0] === this.transformInfo ? null : decorator[0]));
            }
            if (this.onApiMethodExtracted) {
                this.onApiMethodExtracted(Object.assign({}, definition, { metadata }));
            }
        }
        return node;
    }
    parseApiMethodCallParameters(p) {
        const parameters = [...p];
        const parsedParams = [];
        for (let i = 0; i < parameters.length; ++i) {
            let param = parameters[i];
            for (const decorator of param.decorators) {
                for (const decoratorType of this.transformInfo.parameterTypes) {
                    if (this.isArgumentDecoratorCallExpression(decorator.expression, decoratorType)) {
                        const transformer = this.paramTransformers.get(decoratorType);
                        const paramArgs = transformer.getParamArgs(param);
                        parsedParams.push({
                            args: transformer.getDecoratorArguments(paramArgs, decorator.expression, decoratorType),
                            parameterIndex: i,
                            propertyKey: paramArgs.name,
                            type: decoratorType.type,
                        });
                        param = transformer.visitNode(param);
                    }
                }
            }
            parameters[i] = param;
        }
        return {
            paramDefs: parsedParams,
            parameters: ts.factory.createNodeArray(parameters),
        };
    }
    isDecoratedMethodDeclaration(node) {
        if (!ts.isMethodDeclaration(node)) {
            return false;
        }
        if (!node.decorators || node.decorators.length === 0) {
            return false;
        }
        for (const decorator of node.decorators) {
            if (this.isArgumentDecoratorCallExpression(decorator.expression)) {
                return true;
            }
        }
        return false;
    }
    getDecoratorArguments(decoratorArg, expression) {
        const definition = Object.assign({}, decoratorArg);
        const exprArguments = [];
        for (let i = 0; i < this.transformInfo.arguments.length; ++i) {
            const argDef = this.transformInfo.arguments[i];
            const arg = expression.arguments[i];
            if (typeof arg === 'undefined') {
                if (argDef.optional) {
                    if (!argDef.transformedParameter) {
                        break;
                    }
                }
                else if (!argDef.optional) {
                    throw new Error('Expected argument');
                }
            }
            switch (argDef.type) {
                case 'route':
                    definition.route = this.typeSerializer.compileExpressionToStringConstant(arg);
                    exprArguments[i] = arg;
                    break;
                case 'returnSchema':
                    if (definition.returnType) {
                        exprArguments[i] = this.typeSerializer.objectToLiteral(definition.returnType);
                    }
                    else {
                        exprArguments[i] = arg;
                    }
                    break;
                default:
                    throw new Error(`Unknown argdef type for decorator "${this.transformInfo.apiDecoratorMethod}": "${argDef.type}"`);
            }
        }
        return { definition, exprArguments: ts.factory.createNodeArray(exprArguments) };
    }
}
exports.ExtractionTransformer = ExtractionTransformer;
//# sourceMappingURL=ExtractionTransformer.js.map