"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HandlerMethodParameterDecorator = void 0;
const ts = require("typescript");
const DecoratorDefinition_1 = require("./DecoratorDefinition");
const HandlerTree_1 = require("./HandlerTree");
const InternalTypes_1 = require("../apiManagement/InternalTypes");
const TransformerMetadata_1 = require("./TransformerMetadata");
const Decorator_1 = require("./Decorator");
const ExpressionWrapper_1 = require("./ExpressionWrapper");
const CompilationError_1 = require("../Util/CompilationError");
class HandlerMethodParameterDecorator extends Decorator_1.Decorator {
    constructor(definition) {
        super({
            ...definition,
            decoratorType: DecoratorDefinition_1.DecoratorType.MethodParameter,
        }, Decorator_1.DecoratorNodeType.Parameter);
    }
    get parameterTypeRestrictions() {
        return this.definition.parameterTypeRestrictions;
    }
    get parameterType() {
        return this.definition.parameterType;
    }
    get transportTypeId() {
        return this.definition.transportTypeId;
    }
    get paramId() {
        return this.definition.paramId;
    }
    get bodyType() {
        return this.definition.bodyType;
    }
    get skipOutputTypeDefinitions() {
        return !!this.definition.skipOutputTypeDefinitions;
    }
    get overrideOutput() {
        return !!this.definition.overrideOutput;
    }
    getDecoratorTreeElement(parent, node, decorator, context) {
        const argumentResult = this.applyArguments(node, decorator, context);
        const decoratorTreeNode = {
            type: HandlerTree_1.HandlerTreeNodeType.HandlerParameter,
            decorator: this.definition,
            children: [],
            parent,
            location: this.getNodeLocation(node),
            paramDef: {
                args: {
                    name: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(argumentResult.metadata, TransformerMetadata_1.BuiltinMetadata.Name),
                    optional: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(argumentResult.metadata, TransformerMetadata_1.BuiltinMetadata.Optional),
                    typedef: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(argumentResult.metadata, TransformerMetadata_1.BuiltinMetadata.Typedef),
                    description: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(argumentResult.metadata, TransformerMetadata_1.BuiltinMetadata.Description),
                    initializer: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(argumentResult.metadata, TransformerMetadata_1.BuiltinMetadata.Initializer),
                    regexp: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(argumentResult.metadata, TransformerMetadata_1.BuiltinMetadata.ValidationRegExp),
                    typeref: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(argumentResult.metadata, TransformerMetadata_1.BuiltinMetadata.Typeref),
                    validationFunc: (0, TransformerMetadata_1.getMetadataValueByDescriptor)(argumentResult.metadata, TransformerMetadata_1.BuiltinMetadata.ValidationFunction),
                },
                parameterIndex: node.parent.parameters.indexOf(node),
                propertyKey: this.getParameterName(node),
                type: this.definition.parameterType,
                transportTypeId: this.transportTypeId,
                paramId: this.paramId,
                bodyType: this.bodyType,
                overrideOutput: this.overrideOutput,
            },
            metadata: argumentResult.metadata,
        };
        decoratorTreeNode.metadata = decoratorTreeNode.metadata.concat(context.metadataManager.getApiMetadataForApiMethodParam(decoratorTreeNode, node, this));
        return {
            decoratorTreeNode,
            transformedDecorator: argumentResult.decorator,
        };
    }
    *getDefaultMetadataGetters() {
        yield* super.getDefaultMetadataGetters();
        yield (node, decorator, context) => this.getNodeTyperefMetadata(node, decorator, context);
        yield (node, decorator, context) => this.getNodeTypedefMetadata(node, decorator, context);
        yield (node, decorator, context) => this.getNodeOptionalMetadata(node, decorator, context);
        yield (node, decorator, context) => this.getNodeInitializerMetadata(node, decorator, context);
        yield (node, decorator, context) => this.getNodeDescriptionMetadata(node, decorator, context);
        yield (node, decorator, context) => this.getNodeParameterIndexMetadata(node, decorator, context);
        yield (node, decorator, context) => this.getNodePropertyKeyMetadata(node, decorator, context);
    }
    getNodeTyperefMetadata(node, decorator, context) {
        if (this.skipOutputTypeDefinitions) {
            return;
        }
        let internalType = {
            type: 'any',
        };
        if (node.type) {
            const type = context.typeChecker.getTypeFromTypeNode(node.type);
            internalType = context.typeSerializer.getInternalTypeRepresentation(node.type, type);
            if (!internalType) {
                internalType = InternalTypes_1.InternalTypeUtil.TypeAny;
            }
            if (this.parameterTypeRestrictions
                && !this.parameterTypeRestrictions.find(t => t.type === internalType.type || t.type === 'any')) {
                throw new CompilationError_1.CompilationError(`Invalid type for decorator '${this.magicFunctionName}': ${internalType.type}`, node);
            }
        }
        return {
            ...TransformerMetadata_1.BuiltinMetadata.Typedef,
            value: internalType,
        };
    }
    getNodeTypedefMetadata(node, decorator, context) {
        if (this.skipOutputTypeDefinitions) {
            return;
        }
        if (node.type) {
            const type = context.typeChecker.getTypeFromTypeNode(node.type);
            let internalType = context.typeSerializer.getInternalTypeRepresentation(node.type, type);
            if (!internalType) {
                internalType = InternalTypes_1.InternalTypeUtil.TypeAny;
            }
            if (this.parameterTypeRestrictions
                && !this.parameterTypeRestrictions.find(t => t.type === internalType.type || t.type === 'any')) {
                throw new CompilationError_1.CompilationError(`Invalid type for decorator '${this.magicFunctionName}': ${internalType.type}`, node);
            }
            if (internalType.type === 'object' && type.isClass() && ts.isTypeReferenceNode(node.type)) {
                if (ts.isIdentifier(node.type.typeName)) {
                    return {
                        ...TransformerMetadata_1.BuiltinMetadata.Typeref,
                        value: new ExpressionWrapper_1.ExpressionWrapper(node.type.typeName),
                    };
                }
            }
        }
    }
    getNodeOptionalMetadata(node, decorator, context) {
        let optional = !!node.questionToken;
        if (optional) {
            return {
                ...TransformerMetadata_1.BuiltinMetadata.Optional,
                value: optional,
            };
        }
    }
    getNodeInitializerMetadata(node, decorator, context) {
        if (node.initializer) {
            let parenExpr = this.parenthesizeExpression(node.initializer);
            return {
                ...TransformerMetadata_1.BuiltinMetadata.Initializer,
                value: new ExpressionWrapper_1.ExpressionWrapper(ts.factory.createArrowFunction(undefined, undefined, [], undefined, undefined, parenExpr)),
            };
        }
    }
    getNodeDescriptionMetadata(node, decorator, context) {
        let description = this.getParamDescription(node);
        if (description) {
            return {
                ...TransformerMetadata_1.BuiltinMetadata.Description,
                value: description,
            };
        }
    }
    getNodeParameterIndexMetadata(node, decorator, context) {
        return {
            ...TransformerMetadata_1.BuiltinMetadata.ParameterIndex,
            value: node.parent.parameters.indexOf(node),
        };
    }
    getNodePropertyKeyMetadata(node, decorator, context) {
        return {
            ...TransformerMetadata_1.BuiltinMetadata.PropertyKey,
            value: this.getParameterName(node),
        };
    }
    getParameterName(node) {
        let name;
        if (typeof node.name === 'string') {
            name = node.name;
        }
        else if (ts.isIdentifier(node.name)) {
            name = node.name.text;
        }
        else {
            throw new Error('Unknown node name type');
        }
        return name;
    }
    getParamDescription(param) {
        const paramTags = ts.getJSDocParameterTags(param);
        if (paramTags.length) {
            return typeof paramTags[0].comment === 'string' ? paramTags[0].comment : paramTags[0].comment?.map(c => c.text).join(' ');
        }
        return undefined;
    }
}
exports.HandlerMethodParameterDecorator = HandlerMethodParameterDecorator;
//# sourceMappingURL=HandlerMethodParameterDecorator.js.map