"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OpenApiMetadataExtractors = void 0;
const ts = require("typescript");
const TransformerUtil_1 = require("./TransformerUtil");
const HandlerTree_1 = require("./HandlerTree");
class OpenApiMetadataExtractors {
    constructor() { }
    ;
    static RegisterMetadataExtractors(mm) {
        mm.addApiMethodMetadataGenerator(this.ExtractMethodJsDocMetadata);
        mm.addApiMethodCollectionMetadataGenerator(this.ExtractClassJsDocMetadata);
    }
    static ExtractClassJsDocMetadata(method, methodNode) {
        const metadata = [];
        if (method.type !== HandlerTree_1.HandlerTreeNodeType.HandlerCollection) {
            return metadata;
        }
        if ((0, TransformerUtil_1.isNodeWithJsDoc)(methodNode) && methodNode.jsDoc.length > 0) {
            // jsdoc tags
            const firstLine = OpenApiMetadataExtractors.getFirstCommentLine(methodNode);
            if (firstLine) {
                OpenApiMetadataExtractors.getJsDocTags(firstLine)
                    .map(t => ({
                    type: "OpenApi" /* IMetadataType.OpenApi */,
                    key: "tag" /* OpenApiMetadataType.Tag */,
                    value: t,
                }))
                    .forEach(t => metadata.push(t));
            }
        }
        return metadata;
    }
    static ExtractMethodJsDocMetadata(method, methodNode) {
        const metadata = [];
        if (method.type !== HandlerTree_1.HandlerTreeNodeType.Handler) {
            return metadata;
        }
        if ((0, TransformerUtil_1.isNodeWithJsDoc)(methodNode) && methodNode.jsDoc.length > 0) {
            // Method description
            metadata.push({
                type: "OpenApi" /* IMetadataType.OpenApi */,
                key: "description" /* OpenApiMetadataType.Description */,
                value: methodNode.jsDoc.reduce((prev, current) => prev + current.comment, ''),
            });
            // Return tag
            const returnTag = OpenApiMetadataExtractors.jsdocTagString(ts.getJSDocReturnTag(methodNode));
            if (returnTag) {
                metadata.push({
                    type: "OpenApi" /* IMetadataType.OpenApi */,
                    key: "response" /* OpenApiMetadataType.ResponseDescription */,
                    value: returnTag,
                });
            }
            // Private tag
            const privateTags = OpenApiMetadataExtractors.getJsDocCustomTags(methodNode.jsDoc, 'private');
            if (privateTags) {
                metadata.push(...privateTags.map(t => ({
                    type: "OpenApi" /* IMetadataType.OpenApi */,
                    key: "private" /* OpenApiMetadataType.Private */,
                    value: t,
                })));
            }
            // Method summary and jsdoc tags
            const firstLine = OpenApiMetadataExtractors.getFirstCommentLine(methodNode);
            if (firstLine) {
                metadata.push({
                    type: "OpenApi" /* IMetadataType.OpenApi */,
                    key: "summary" /* OpenApiMetadataType.Summary */,
                    value: firstLine.comment,
                });
                OpenApiMetadataExtractors.getJsDocTags(firstLine)
                    .map(t => ({
                    type: "OpenApi" /* IMetadataType.OpenApi */,
                    key: "tag" /* OpenApiMetadataType.Tag */,
                    value: t,
                }))
                    .forEach(t => metadata.push(t));
            }
        }
        return metadata;
    }
    static getFirstCommentLine(methodNode) {
        return methodNode.jsDoc.find(n => n && n.comment && n.comment.length > 0);
    }
    static getJsDocTags(node) {
        if (node.tags) {
            return (node.tags
                .filter(t => t.tagName.text === OpenApiMetadataExtractors.JsDocTagTag)
                .map(t => {
                const parts = typeof t.comment === 'string' ? t.comment?.split(/\s+/) : t.comment?.map(c => c.text);
                return {
                    name: parts.shift(),
                    description: parts.join(' '),
                };
            }));
        }
        return [];
    }
    static getJsDocCustomTags(nodes, tag) {
        return nodes.map(node => {
            if (node.tags) {
                return node.tags.filter(t => t.tagName.text === tag)
                    .map(t => {
                    const parts = typeof t.comment === 'string' ? t.comment?.split(/\s+/) : t.comment?.map(c => c.text);
                    return {
                        name: parts ? parts.shift() : tag,
                        description: parts?.join(' '),
                    };
                });
            }
            else {
                return [];
            }
        }).reduce((p, c) => p.concat(c), []);
    }
    static jsdocTagString(tag) {
        if (tag) {
            return typeof tag.comment === 'string' ? tag.comment : tag.comment?.map(c => c.text).join(' ');
        }
        return undefined;
    }
}
exports.OpenApiMetadataExtractors = OpenApiMetadataExtractors;
OpenApiMetadataExtractors.JsDocTagTag = 'tags';
//# sourceMappingURL=OpenApi.js.map