"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const DecoratorTransformer_1 = require("./DecoratorTransformer");
class ParamDecoratorTransformer extends DecoratorTransformer_1.DecoratorTransformer {
    constructor(program, generator, transformInfo, extractOnly = false) {
        super(program, generator, Object.assign({}, transformInfo, { nodeCheckFunction: (node => this.isDecoratedParameterExpression(node)) }));
        this.extractOnly = extractOnly;
    }
    visitNode(node) {
        // TODO: Validate the type of the parameter this decorator was used on matches the allowable types
        // defined on: this.transformInfo.allowableTypes
        // Parse type
        const decoratorArg = this.getParamArgs(node);
        // Replace decorator definitions
        for (const decorator of node.decorators) {
            if (this.isArgumentDecoratorCallExpression(decorator.expression)) {
                // Replace decorator invocation
                const thisArgs = this.getDecoratorArguments(decoratorArg, decorator.expression);
                if (thisArgs && !this.extractOnly) {
                    // Change decorator arguments
                    decorator.expression.arguments = ts.factory.createNodeArray([this.typeSerializer.objectToLiteral(thisArgs)]);
                }
            }
        }
        return node;
    }
    getParamArgs(node) {
        let internalType = {
            type: 'any',
        };
        let typeref = null;
        if (node.type) {
            const type = this.typeChecker.getTypeFromTypeNode(node.type);
            internalType = this.typeSerializer.getInternalTypeRepresentation(node.type, type);
            if (!this.transformInfo.allowableTypes.find(t => t === internalType.type || t === 'any')) {
                throw new Error('Invalid type for decorator: ' + internalType.type);
            }
            if (internalType.type === 'object' && type.isClass() && ts.isTypeReferenceNode(node.type)) {
                if (ts.isIdentifier(node.type.typeName)) {
                    typeref = new DecoratorTransformer_1.ExpressionWrapper(node.type.typeName);
                }
            }
        }
        // Parse argument name
        let name;
        if (typeof node.name === 'string') {
            name = node.name;
        }
        else if (ts.isIdentifier(node.name)) {
            name = node.name.text;
        }
        else {
            throw new Error('Unknown node name type');
        }
        // Parse initializer
        let initializer;
        if (node.initializer) {
            // Only wrap the expression if it's not already parenthesized
            let parenExpr = this.parenthesizeExpression(node.initializer);
            initializer = ts.factory.createArrowFunction(undefined, undefined, [], undefined, undefined, parenExpr);
        }
        // Parse optional
        let optional = !!node.questionToken;
        // Construct param decorator args
        let description;
        const args = Object.assign({ name, typedef: internalType }, ((description = this.getParamDescription(node))
            ? { description }
            : {}));
        if (optional) {
            args.optional = optional;
        }
        if (typeref) {
            args.typeref = typeref;
        }
        const decoratorArg = Object.assign({}, args, (initializer
            ? { initializer: new DecoratorTransformer_1.ExpressionWrapper(initializer) }
            : {}));
        return decoratorArg;
    }
    getParamDescription(param) {
        const paramTags = ts.getJSDocParameterTags(param);
        if (paramTags.length) {
            return paramTags[0].comment;
        }
        return undefined;
    }
    getDecoratorArguments(decoratorArg, expression, transformInfo = this.transformInfo) {
        const thisArgs = Object.assign({}, decoratorArg);
        for (let i = 0; i < transformInfo.arguments.length; ++i) {
            const argDef = transformInfo.arguments[i];
            const arg = expression.arguments[i];
            if (this.isNullOrUndefinedExpr(arg)) {
                if (!argDef.optional) {
                    throw new Error('Expected argument');
                }
                continue;
            }
            switch (argDef.type) {
                case 'paramName':
                    thisArgs.name = this.typeSerializer.compileExpressionToStringConstant(arg);
                    break;
                case 'validationFunc':
                    thisArgs.validationFunction = new DecoratorTransformer_1.ExpressionWrapper(this.parenthesizeExpression(arg));
                    break;
                case 'numberMax':
                    thisArgs.typedef.maxVal = this.typeSerializer.compileExpressionToNumericConstant(arg);
                    break;
                case 'numberMin':
                    thisArgs.typedef.minVal = this.typeSerializer.compileExpressionToNumericConstant(arg);
                    break;
                case 'regexp':
                    if (ts.isRegularExpressionLiteral(arg)) {
                        thisArgs.regexp = new DecoratorTransformer_1.ExpressionWrapper(arg);
                    }
                    else {
                        throw new Error('Regular expression must be a regex literal');
                    }
                    break;
                default:
                    throw new Error('Unknown argdef type');
            }
        }
        return thisArgs;
    }
    isNullOrUndefinedExpr(arg) {
        if (typeof arg === 'undefined') {
            return true;
        }
        else if (ts.isParenthesizedExpression(arg)) {
            return this.isNullOrUndefinedExpr(arg.expression);
        }
        else if (arg.kind === ts.SyntaxKind.UndefinedKeyword || arg.kind === ts.SyntaxKind.NullKeyword) {
            return true;
        }
        else if (ts.isIdentifier(arg)) {
            return arg.originalKeywordKind === ts.SyntaxKind.UndefinedKeyword || arg.originalKeywordKind === ts.SyntaxKind.NullKeyword;
        }
        else {
            return false;
        }
    }
}
exports.ParamDecoratorTransformer = ParamDecoratorTransformer;
//# sourceMappingURL=ParamDecoratorTransformer.js.map