"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TreeTransformer = void 0;
const ts = require("typescript");
const TypeSerializer_1 = require("./TypeSerializer");
const Decorator_1 = require("./Decorator");
const HandlerTree_1 = require("./HandlerTree");
const DecoratorDefinition_1 = require("./DecoratorDefinition");
class TreeTransformer {
    constructor(program, generator, decorators, applyTransformation, metadataManager) {
        this.program = program;
        this.generator = generator;
        this.decorators = decorators;
        this.applyTransformation = applyTransformation;
        this.transformedNodes = new Map();
        this.rootNode = {
            type: HandlerTree_1.HandlerTreeNodeType.Root,
            children: [],
            metadata: [],
        };
        this.typeChecker = this.program.getTypeChecker();
        this.typeSerializer = new TypeSerializer_1.TypeSerializer(program, generator, this.typeChecker);
        this.transformContext = {
            generator,
            metadataManager,
            program,
            typeChecker: this.typeChecker,
            typeSerializer: this.typeSerializer,
        };
    }
    get root() {
        return this.rootNode;
    }
    visitNode(node, context) {
        if (this.isPotentialClassContainer(node)) {
            return this.visitNodeChildren(node, context);
        }
        return this.visitNodeInTreeContext(node, context, this.rootNode);
    }
    visitNodeChildren(node, context) {
        return ts.visitEachChild(node, node => this.visitNode(node, context), context);
    }
    visitNodeInTreeContext(node, context, parent) {
        const nodeType = this.getTypeForNode(node);
        if (typeof nodeType === 'undefined' || !this.nodeHasDecorators(node)) {
            if (parent) {
                return this.visitNodeChildrenInTreeContext(node, context, parent);
            }
            return this.visitNodeChildren(node, context);
        }
        let nodeWithDecorators = node;
        const decorators = this.decorators.getDecoratorsForNodeType(nodeType, parent);
        const nodeDecorators = [];
        const nodeTypes = {
            [DecoratorDefinition_1.DecoratorNodeTreeHierarchyType.Child]: [],
            [DecoratorDefinition_1.DecoratorNodeTreeHierarchyType.Modifier]: [],
        };
        for (let i = 0; i < nodeWithDecorators.modifiers.length; ++i) {
            const modifier = nodeWithDecorators.modifiers[i];
            if (!ts.isDecorator(modifier)) {
                nodeDecorators[i] = undefined;
                continue;
            }
            nodeDecorators[i] = modifier;
            for (const definition of decorators) {
                if ((definition.isCallExpression && this.isArgumentDecoratorCallExpression(nodeDecorators[i].expression, definition))
                    || (!definition.isCallExpression && this.isArgumentDecoratorExpression(nodeDecorators[i].expression, definition))) {
                    const treeNode = this.getDecoratorTreeElement(definition, parent, node, nodeDecorators[i]);
                    nodeTypes[definition.treeHierarchyType].push(treeNode.decoratorTreeNode);
                    if (treeNode.transformedDecorator) {
                        nodeDecorators[i] = treeNode.transformedDecorator;
                        this.cacheDecoratorTreeElement(node, nodeDecorators[i], treeNode);
                    }
                    nodeWithDecorators = this.visitNodeChildrenInTreeContext(node, context, treeNode.decoratorTreeNode);
                }
            }
        }
        for (const node of nodeTypes[DecoratorDefinition_1.DecoratorNodeTreeHierarchyType.Child]) {
            parent.children.push(node);
            for (const modifierNode of nodeTypes[DecoratorDefinition_1.DecoratorNodeTreeHierarchyType.Modifier]) {
                node.children.push(modifierNode);
            }
        }
        if (this.applyTransformation) {
            nodeWithDecorators = {
                ...nodeWithDecorators,
                modifiers: ts.factory.createNodeArray(nodeDecorators.map((d, i) => d ?? nodeWithDecorators.modifiers[i])),
            };
        }
        return nodeWithDecorators;
    }
    getDecoratorTreeElement(definition, parent, node, decorator) {
        if (!this.transformedNodes.has(node)) {
            this.transformedNodes.set(node, new Map());
        }
        const set = this.transformedNodes.get(node);
        if (!set.has(decorator)) {
            this.cacheDecoratorTreeElement(node, decorator, definition.getDecoratorTreeElement(parent, node, decorator, this.transformContext));
        }
        return set.get(decorator);
    }
    cacheDecoratorTreeElement(node, decorator, element) {
        if (!this.transformedNodes.has(node)) {
            this.transformedNodes.set(node, new Map());
        }
        const set = this.transformedNodes.get(node);
        if (!set.has(decorator)) {
            set.set(decorator, element);
        }
    }
    visitNodeChildrenInTreeContext(node, context, parent) {
        return ts.visitEachChild(node, node => this.visitNodeInTreeContext(node, context, parent), context);
    }
    nodeHasDecorators(node) {
        return (ts.canHaveDecorators(node)
            && node.modifiers
            && node.modifiers.some(m => m.kind === ts.SyntaxKind.Decorator));
    }
    /**
     * Returns true if this node type could potentially contain a class
     * @param node
     */
    isPotentialClassContainer(node) {
        return ts.isSourceFile(node) || ts.isModuleBlock(node) || ts.isModuleDeclaration(node) || ts.isBlock(node);
    }
    getTypeForNode(node) {
        if (ts.isMethodDeclaration(node)) {
            return Decorator_1.DecoratorNodeType.Method;
        }
        else if (ts.isParameter(node)) {
            return Decorator_1.DecoratorNodeType.Parameter;
        }
        else if (ts.isPropertyDeclaration(node)) {
            return Decorator_1.DecoratorNodeType.ClassProperty;
        }
        else if (ts.isClassDeclaration(node)) {
            return Decorator_1.DecoratorNodeType.Class;
        }
    }
    isArgumentDecoratorCallExpression(decoratorExpression, definition) {
        if (!ts.isCallExpression(decoratorExpression)) {
            return false;
        }
        const signature = this.typeChecker.getResolvedSignature(decoratorExpression);
        if (typeof signature === 'undefined') {
            return false;
        }
        return this.isMatchingDeclaration(signature.declaration, definition);
    }
    isMatchingDeclaration(declaration, definition) {
        if (!(!!declaration
            && !ts.isJSDocSignature(declaration)
            && !!declaration.name
            && declaration.name.getText() === definition.magicFunctionName)) {
            return false;
        }
        return definition.isSourceFileMatch(declaration.getSourceFile());
    }
    isArgumentDecoratorExpression(decoratorExpression, definition) {
        if (ts.isCallExpression(decoratorExpression) || !ts.isIdentifier(decoratorExpression)) {
            return false;
        }
        // const symbol = this.typeChecker.getSymbolAtLocation(decoratorExpression);
        // if (!symbol.declarations || symbol.declarations.length === 0) {
        // 	return false;
        // }
        // let declaration: ts.NamedDeclaration;
        // for (const decl of symbol.declarations) {
        // 	if (isNamedDeclaration(decl)) {
        // 		declaration = decl;
        // 		break;
        // 	}
        // }
        // if (!declaration) {
        // 	return false;
        // }
        const signatures = [
            ...this.typeChecker.getSignaturesOfType(this.typeChecker.getTypeAtLocation(decoratorExpression), ts.SignatureKind.Call),
            ...this.typeChecker.getSignaturesOfType(this.typeChecker.getTypeAtLocation(decoratorExpression), ts.SignatureKind.Construct),
        ];
        return signatures.some(signature => {
            if (!signature || !signature.declaration) {
                return false;
            }
            return this.isMatchingDeclaration(signature.declaration, definition);
        });
    }
}
exports.TreeTransformer = TreeTransformer;
//# sourceMappingURL=TreeTransformer.js.map