"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TypeSerializer = void 0;
const ts = require("typescript");
const InternalTypes_1 = require("../apiManagement/InternalTypes");
const TransformerUtil_1 = require("./TransformerUtil");
const ExpressionWrapper_1 = require("./ExpressionWrapper");
class TypeSerializer {
    constructor(program, generator, typeChecker) {
        this.program = program;
        this.generator = generator;
        this.typeChecker = typeChecker;
    }
    getInternalTypeRepresentation(node, type) {
        const base = {
            ...(node && ts.isPropertySignature(node)
                ? { optional: !!node.questionToken }
                : {})
        };
        if ((0, TransformerUtil_1.isIntrinsicType)(type)) {
            return {
                ...base,
                type: type.intrinsicName,
            };
        }
        if (node && (ts.isRegularExpressionLiteral(node) || node.getText() === RegExp.name)) {
            return {
                ...base,
                type: 'regex',
            };
        }
        if (type.symbol && (0, TransformerUtil_1.isBuiltinSymbol)(type.symbol)) {
            switch (type.symbol.name) {
                case 'Buffer':
                    return {
                        ...base,
                        type: 'Buffer'
                    };
                case 'Promise':
                    if ((0, TransformerUtil_1.isParameterizedType)(type)) {
                        if ((0, TransformerUtil_1.isNodeWithTypeArguments)(node)) {
                            return {
                                ...base,
                                ...this.getInternalTypeRepresentation(node.typeArguments[0], type.resolvedTypeArguments[0]),
                            };
                        }
                        return {
                            ...base,
                            ...this.getInternalTypeRepresentation(node, type.resolvedTypeArguments[0]),
                        };
                    }
                    else {
                        return {
                            ...base,
                            type: 'Promise'
                        };
                    }
            }
        }
        if (type.symbol && (0, TransformerUtil_1.isSymbolWithId)(type.symbol)) {
            const name = this.typeChecker.getFullyQualifiedName(type.symbol);
            const symbols = [
                ...this.generator.getSymbols(name),
                ...this.generator.getSymbols(type.symbol.name),
            ];
            if ((0, TransformerUtil_1.isSymbolWithParent)(type.symbol)) {
                symbols.push(...this.generator.getSymbols(type.symbol.parent.name));
            }
            for (const symbol of symbols) {
                if ((0, TransformerUtil_1.isSymbolWithId)(symbol.symbol) && (symbol.symbol.id === type.symbol.id
                    || (0, TransformerUtil_1.isSymbolWithParent)(type.symbol) && (0, TransformerUtil_1.isSymbolWithId)(type.symbol.parent) && symbol.symbol.id === type.symbol.parent.id)) {
                    return this.getTypeForSymbolWithId(symbol, base);
                }
            }
        }
        if (node && ts.isUnionTypeNode(node) && (0, TransformerUtil_1.isUnionType)(node, type)) {
            return this.getUnionType(base, type, node);
        }
        if (node && ts.isIntersectionTypeNode(node) && (0, TransformerUtil_1.isIntersectionType)(node, type)) {
            return {
                ...base,
                type: 'intersection',
                types: type.types.map((t, i) => this.getInternalTypeRepresentation(node.types[i], t)),
                typename: type.aliasSymbol ? type.aliasSymbol.escapedName.toString() : undefined,
                uniqueTypename: type.aliasSymbol ? type.aliasSymbol.name : undefined,
            };
        }
        if (node && node && ts.isArrayTypeNode(node)) {
            return {
                ...base,
                type: 'array',
                elementType: this.getInternalTypeRepresentation(node.elementType, this.typeChecker.getTypeFromTypeNode(node.elementType)),
            };
        }
        if (type.aliasSymbol && type.aliasSymbol.declarations.length > 0) {
            for (const decl of type.aliasSymbol.declarations) {
                if (!ts.isTypeAliasDeclaration(decl)) {
                    continue;
                }
                if (decl.type && ts.isTypeNode(decl.type)) {
                    return this.getInternalTypeRepresentation(decl.type, this.typeChecker.getTypeFromTypeNode(decl.type));
                }
            }
        }
        if (node && ts.isParenthesizedTypeNode(node)) {
            return this.getInternalTypeRepresentation(node.type, this.typeChecker.getTypeFromTypeNode(node.type));
        }
        if (node && ts.isLiteralTypeNode(node)) {
            switch (node.literal.kind) {
                case ts.SyntaxKind.StringLiteral:
                    return { type: 'string', schema: { enum: [node.literal.text] } };
                case ts.SyntaxKind.NumericLiteral:
                    return { type: 'number', schema: { enum: [Number(node.literal.text)] } };
                case ts.SyntaxKind.UndefinedKeyword:
                    return { type: 'void' };
                default:
                    throw new Error(`Cannot convert literal type: ${node.literal.kind}`);
            }
        }
    }
    getUnionType(base, type, node) {
        const union = {
            ...base,
            type: 'union',
            types: type.types
                .map((t, i) => this.getInternalTypeRepresentation(node.types[i], t))
                .filter(t => !!t),
            typename: type.aliasSymbol ? type.aliasSymbol.escapedName.toString() : undefined,
            uniqueTypename: type.aliasSymbol ? type.aliasSymbol.name : undefined,
        };
        if (union.types && union.types.every(t => (t.type === 'string' || t.type === 'number') && t.schema && 'enum' in t.schema && Array.isArray(t.schema?.enum))) {
            // This union can be collapsed to an enum
            let type = 'enum';
            if (union.types.every(t => t.type === 'string')) {
                type = 'string';
            }
            else if (union.types.every(t => t.type === 'number')) {
                type = 'number';
            }
            return {
                ...base,
                type,
                schema: {
                    enum: union.types.reduce((p, c) => p.concat('enum' in c.schema ? c.schema.enum : []), []),
                },
                typename: union.typename,
                uniqueTypename: union.uniqueTypename,
            };
        }
        return union;
    }
    getTypeForSymbolWithId(symbol, base) {
        let schema = this.generator.getSchemaForSymbol(symbol.name, true);
        let type = schema.type || InternalTypes_1.InternalTypeUtil.TypeAnyObject.type;
        if (schema.$ref && schema.definitions) {
            const refDefParts = schema.$ref.split('/');
            const refDefName = refDefParts[refDefParts.length - 1];
            const def = schema.definitions[refDefName];
            if (typeof def === 'boolean') {
                throw new Error('Boolean definition: ' + refDefName);
            }
            if (def.type !== InternalTypes_1.InternalTypeUtil.TypeAnyObject.type) {
                if (def.enum) {
                    type = (def.type && !Array.isArray(def.type)) ? def.type : InternalTypes_1.InternalTypeUtil.TypeEnum.type;
                    schema = {
                        ...def,
                        ...schema,
                    };
                }
                else {
                    type = def.type;
                    schema = def;
                }
            }
        }
        return {
            ...base,
            type,
            schema: this.reduceSchemaReferences(schema),
            typename: symbol.typeName,
            uniqueTypename: symbol.name,
        };
    }
    reduceSchemaReferences(schema) {
        if (!schema.$ref || !schema.definitions) {
            return schema;
        }
        const newSchema = {
            ...schema,
            definitions: {},
        };
        const refs = [this.getRefShortName(schema.$ref)];
        const def = schema.definitions[refs[0]];
        if (typeof def === 'boolean') {
            throw new Error('Boolean definition: ' + schema.$ref);
        }
        this.getRefsRecursive(def, schema.definitions, refs);
        refs.forEach(v => newSchema.definitions[v] = schema.definitions[v]);
        return newSchema;
    }
    getRefsRecursive(obj, definitions, refs) {
        Object.keys(obj).filter(k => typeof obj[k] === 'object')
            .map(k => this.getRefsRecursive(obj[k], definitions, refs));
        if (obj.$ref) {
            const ref = this.getRefShortName(obj.$ref);
            if (refs.indexOf(ref) !== -1) {
                return;
            }
            refs.push(ref);
            const definition = definitions[ref];
            if (definition) {
                this.getRefsRecursive(definition, definitions, refs);
            }
        }
    }
    getRefShortName(ref) {
        ref = ref.substring('#/definitions/'.length);
        if (ref.indexOf('/')) {
            ref = ref.split('/')[0];
        }
        return ref;
    }
    getReferencedTypeDefinition(reference) {
        if (!reference.$ref) {
            throw new Error('Reference did not have a $ref field');
        }
        let def = this.generator.ReffedDefinitions[reference.$ref]
            || this.generator.ReffedDefinitions[reference.$ref.substr(TypeSerializer.ReferencePreamble.length)];
        if (!def) {
            throw new Error('Reference not found: ' + reference.$ref);
        }
        // TODO: This is a hack to get around the fact that the generator have the same json schema def
        return def;
    }
    valueToLiteral(val) {
        switch (typeof val) {
            case 'string':
                return ts.factory.createStringLiteral(val);
            case 'number':
                return ts.factory.createNumericLiteral(val);
            case 'boolean':
                return val ? ts.factory.createTrue() : ts.factory.createFalse();
            case 'bigint':
                return ts.factory.createBigIntLiteral(val.toString());
            case 'object':
                if (Array.isArray(val)) {
                    return ts.factory.createArrayLiteralExpression(val.map(v => this.valueToLiteral(v)));
                }
                else if (val instanceof ExpressionWrapper_1.ExpressionWrapper) {
                    return val.node;
                }
                else {
                    return this.objectToLiteral(val);
                }
            case 'undefined':
                return ts.factory.createIdentifier('undefined');
            default:
                throw new Error(`Unknown type serialization path: ${typeof val}`);
        }
    }
    objectToLiteral(val) {
        return ts.factory.createObjectLiteralExpression(Object.keys(val).map(k => {
            let propName = k;
            if (/[^a-zA-Z0-9_]/.test(k)) {
                propName = ts.factory.createStringLiteral(k);
            }
            return ts.factory.createPropertyAssignment(propName, this.valueToLiteral(val[k]));
        }), false);
    }
    compileExpressionToNumericConstant(expression) {
        const compiled = this.compileExpressionToConstant(expression);
        if (typeof compiled === 'number' || typeof compiled === 'bigint') {
            return compiled;
        }
        else if (typeof compiled === 'string') {
            return Number(compiled);
        }
        else if (typeof compiled === 'undefined') {
            return undefined;
        }
        else if (compiled === null) {
            return null;
        }
        else {
            throw new Error('Unknown compilation result type: ' + typeof compiled);
        }
    }
    compileExpressionToBooleanConstant(expression) {
        return !!this.compileExpressionToConstant(expression);
    }
    compileExpressionToStringConstant(expression) {
        const compiled = this.compileExpressionToConstant(expression);
        if (typeof compiled === 'number' || typeof compiled === 'bigint') {
            return compiled.toString();
        }
        else if (typeof compiled === 'string') {
            return compiled;
        }
        else if (typeof compiled === 'undefined') {
            return undefined;
        }
        else if (compiled === null) {
            return null;
        }
        else {
            throw new Error('Unknown compilation result type: ' + typeof compiled);
        }
    }
    compileExpressionToArrayConstant(expression) {
        if (expression.kind !== ts.SyntaxKind.ArrayLiteralExpression) {
            throw new Error('Expression was of invalid kind to compile to array: ' + expression.kind);
        }
        if ((0, TransformerUtil_1.nodeHasElements)(expression)) {
            return expression.elements.map(c => this.arrayNodeToConstant(c)).filter(c => Boolean(c));
        }
        try {
            return expression.getChildren().map(c => this.arrayNodeToConstant(c)).filter(c => Boolean(c));
        }
        catch (e) {
        }
        return expression.forEachChild(undefined, (children) => children.map(c => this.arrayNodeToConstant(c))).filter(c => Boolean(c));
    }
    arrayNodeToConstant(c) {
        if (!ts.isLiteralExpression(c)) {
            if (ts.isToken(c)) {
                return null;
            }
            throw new Error('Array elements must be literals');
        }
        if (c.kind === ts.SyntaxKind.SyntaxList && c.getChildCount() === 1) {
            return this.compileExpressionToConstant(c.getChildAt(0));
        }
        return this.compileExpressionToConstant(c);
    }
    compileExpressionToConstant(expression) {
        if (ts.isLiteralExpression(expression)) {
            switch (expression.kind) {
                case ts.SyntaxKind.StringLiteral:
                    return expression.text;
                case ts.SyntaxKind.TrueKeyword:
                    return true;
                case ts.SyntaxKind.FalseKeyword:
                    return false;
                case ts.SyntaxKind.NumericLiteral:
                    return Number(expression.text);
                case ts.SyntaxKind.UndefinedKeyword:
                    return undefined;
                case ts.SyntaxKind.NullKeyword:
                    return null;
                default:
                    throw new Error(`Unknown literal type: ${expression.kind}`);
            }
        }
        if (ts.isEnumMember(expression) || ts.isElementAccessExpression(expression) || ts.isPropertyAccessExpression(expression)) {
            return this.typeChecker.getConstantValue(expression);
        }
        if (ts.isNoSubstitutionTemplateLiteral(expression)) {
            return expression.text;
        }
        if (ts.isTemplateExpression(expression)) {
            let stringParts = [];
            if (expression.head) {
                stringParts.push(this.compileTemplateTextPart(expression.head));
            }
            for (const span of expression.templateSpans) {
                stringParts.push(this.compileTemplateSpan(span));
            }
            return stringParts.join();
        }
    }
    compileTemplateSpan(span) {
        let expression = '';
        let literal = '';
        if (span.expression) {
            expression = this.compileExpressionToConstant(span.expression).toString();
        }
        if (span.literal) {
            if (ts.isTemplateTail(span.literal)) {
                literal = this.compileTemplateTextPart(span.literal);
            }
            else if (ts.isTemplateMiddle(span.literal)) {
                literal = this.compileTemplateTextPart(span.literal);
            }
            else {
                throw new Error(`Unknown template literal type`);
            }
        }
        return expression + literal;
    }
    compileTemplateTextPart(part) {
        return part.text;
    }
}
exports.TypeSerializer = TypeSerializer;
TypeSerializer.ReferencePreamble = '#/definitions/';
//# sourceMappingURL=TypeSerializer.js.map